#!/usr/bin/python

import unittest
import __init__ as test

network = False
proteome = False
proteinParser = False
uniProtAPI = False
proteinObject = False
isoformAlgo = True
ALL = False



if uniProtAPI or ALL:
    suite = unittest.TestLoader().loadTestsFromTestCase(test.UniprotAPI_tests.TestUniprotAPIFunctions)
    unittest.TextTestRunner(verbosity=2).run(suite)

if network or ALL:
    suite = unittest.TestLoader().loadTestsFromTestCase(test.Networking_tests.TestNetworkingFunctions)
    unittest.TextTestRunner(verbosity=2).run(suite)

if proteome or ALL:
    suite = unittest.TestLoader().loadTestsFromTestCase(test.Proteome_tests.TestProteomeFunctions)
    unittest.TextTestRunner(verbosity=2).run(suite)

if proteinParser or ALL:
    suite = unittest.TestLoader().loadTestsFromTestCase(test.ProteinParser_tests.TestProteinParserFunctions)
    unittest.TextTestRunner(verbosity=2).run(suite)

if proteinObject or ALL:
    suite = unittest.TestLoader().loadTestsFromTestCase(test.ProteinObject_tests.TestProteinObjectFunctions)
    unittest.TextTestRunner(verbosity=2).run(suite)

if isoformAlgo or ALL:
    suite = unittest.TestLoader().loadTestsFromTestCase(test.Isoform_algorithm_tests.TestIsoformAlgorithm)
    unittest.TextTestRunner(verbosity=2).run(suite)




import unittest
from geeneus import Proteome

class TestProteomeFunctions(unittest.TestCase):

    # Build manager object for all tests here
    def setUp(self):
        self.manager_cacheOn = Proteome.ProteinManager("alex.holehouse@gmail.com", cache=True)
        self.manager_cacheOff = Proteome.ProteinManager("alex.holehouse@gmail.com", cache=False)


    def test_NCBI_code_coverage(self):
        acc = "729143" # GI for p21 MAPK

        self.manager_cacheOn.get_ID_type(acc)
        self.manager_cacheOn.get_domains(acc)
        self.manager_cacheOn.get_geneID(acc)
        self.manager_cacheOn.get_gene_name(acc)
        self.manager_cacheOn.get_host(acc)
        self.manager_cacheOn.get_isoforms(acc)
        self.manager_cacheOn.get_other_accessions(acc)
        self.manager_cacheOn.get_protein_name(acc)
        self.manager_cacheOn.get_protein_sequence(acc)
        self.manager_cacheOn.get_protein_sequence_length(acc)
        self.manager_cacheOn.get_raw_xml(acc)
        self.manager_cacheOn.get_record_creation_date(acc)
        self.manager_cacheOn.get_record_source(acc)
        self.manager_cacheOn.get_record_version(acc)
        self.manager_cacheOn.get_species(acc)
        self.manager_cacheOn.get_taxonomy(acc)
        self.manager_cacheOn.get_variants(acc)
        
        self.manager_cacheOff.get_ID_type(acc)
        self.manager_cacheOff.get_domains(acc)
        self.manager_cacheOff.get_geneID(acc)
        self.manager_cacheOff.get_gene_name(acc)
        self.manager_cacheOff.get_host(acc)
        self.manager_cacheOff.get_isoforms(acc)
        self.manager_cacheOff.get_other_accessions(acc)
        self.manager_cacheOff.get_protein_name(acc)
        self.manager_cacheOff.get_protein_sequence(acc)
        self.manager_cacheOff.get_protein_sequence_length(acc)
        self.manager_cacheOff.get_raw_xml(acc)
        self.manager_cacheOff.get_record_creation_date(acc)
        self.manager_cacheOff.get_record_source(acc)
        self.manager_cacheOff.get_record_version(acc)
        self.manager_cacheOff.get_species(acc)
        self.manager_cacheOff.get_taxonomy(acc)
        self.manager_cacheOff.get_variants(acc)


    def test_UNIPROT_code_coverage(self):
        acc = "P38936" # GI for p21 MAPK
        
        self.manager_cacheOn.get_ID_type(acc)
        self.manager_cacheOn.get_domains(acc)
        self.manager_cacheOn.get_geneID(acc)
        self.manager_cacheOn.get_gene_name(acc)
        self.manager_cacheOn.get_host(acc)
        self.manager_cacheOn.get_isoforms(acc)
        self.manager_cacheOn.get_other_accessions(acc)
        self.manager_cacheOn.get_protein_name(acc)
        self.manager_cacheOn.get_protein_sequence(acc)
        self.manager_cacheOn.get_protein_sequence_length(acc)
        self.manager_cacheOn.get_raw_xml(acc)
        self.manager_cacheOn.get_record_creation_date(acc)
        self.manager_cacheOn.get_record_source(acc)
        self.manager_cacheOn.get_record_version(acc)
        self.manager_cacheOn.get_species(acc)
        self.manager_cacheOn.get_taxonomy(acc)
        self.manager_cacheOn.get_variants(acc)
        
        self.manager_cacheOff.get_ID_type(acc)
        self.manager_cacheOff.get_domains(acc)
        self.manager_cacheOff.get_geneID(acc)
        self.manager_cacheOff.get_gene_name(acc)
        self.manager_cacheOff.get_host(acc)
        self.manager_cacheOff.get_isoforms(acc)
        self.manager_cacheOff.get_other_accessions(acc)
        self.manager_cacheOff.get_protein_name(acc)
        self.manager_cacheOff.get_protein_sequence(acc)
        self.manager_cacheOff.get_protein_sequence_length(acc)
        self.manager_cacheOff.get_raw_xml(acc)
        self.manager_cacheOff.get_record_creation_date(acc)
        self.manager_cacheOff.get_record_source(acc)
        self.manager_cacheOff.get_record_version(acc)
        self.manager_cacheOff.get_species(acc)
        self.manager_cacheOff.get_taxonomy(acc)
        self.manager_cacheOff.get_variants(acc)



    def test_get_sequence(self):
        # make sure we can pull a protein sequence
     
        testSeq = "maaaaaagagpemvrgqvfdvgprytnlsyigegaygmvcsaydnvnkvrvaikkispfehqtycqrtlreikillrfrheniigindiiraptieqmkdvyivqdlmetdlykllktqhlsndhicyflyqilrglkyihsanvlhrdlkpsnlllnttcdlkicdfglarvadpdhdhtgflteyvatrwyrapeimlnskgytksidiwsvgcilaemlsnrpifpgkhyldqlnhilgilgspsqedlnciinlkarnyllslphknkvpwnrlfpnadskaldlldkmltfnphkrieveqalahpyleqyydpsdepiaeapfkfdmelddlpkeklkelifeetarfqpgyrs"
        
        self.assertEqual(testSeq, self.manager_cacheOff.get_protein_sequence("NP_002736"))
        self.assertEqual(testSeq, self.manager_cacheOn.get_protein_sequence("NP_002736"))
        self.assertEqual(None, self.manager_cacheOff.get_protein_sequence("SHOULDN'T WORK"))
        self.assertEqual(None, self.manager_cacheOn.get_protein_sequence("SHOULDN'T WORK"))

        
    # check the pdb translation is working OK (uses eSearch)
    def test_PDB_translation(self):
        testSeq = "mahhhhhhmaksglrqdpqstaaatvlkraveldsesrypqalvcyqegidlllqvlkgtkdntkrcnlrekiskymdraenikkyldqekedgkyhkqikieenatgfsyeslfreylnetvtevwiedpyirhthqlynflrfcemlikrpckvktihlltsldegieqvqqsrglqeieeslrshgvllevqysssihdreirfnngwmikigrgldyfkkpqsrfslgycdfdlrpchettvdifhkkhtkni"
       
        self.assertEqual(testSeq, self.manager_cacheOff.get_protein_sequence("2YMB_A"))
        self.assertEqual(testSeq, self.manager_cacheOn.get_protein_sequence("2YMB_A"))
       
        print "The following PDB values are generic, so will return a list of GIs which we can't resolve"
        self.assertEqual(None, self.manager_cacheOff.get_protein_sequence("2YMB"))
        self.assertEqual(None, self.manager_cacheOn.get_protein_sequence("2YMB"))

    def test_get_protein_name(self):
        self.assertEqual('p21 [Homo sapiens]', self.manager_cacheOff.get_protein_name("AAB29246"))
        self.assertEqual('p21 [Homo sapiens]', self.manager_cacheOn.get_protein_name("AAB29246"))

    def test_get_raw_xml(self):
        self.assertEqual("AA", self.manager_cacheOff.get_raw_xml("AAB29246")["GBSeq_moltype"])
        self.assertEqual("AA", self.manager_cacheOn.get_raw_xml("AAB29246")["GBSeq_moltype"])

    def test_get_variants(self):
        self.assertEqual(100, self.manager_cacheOff.get_variants("P42685")[0]["location"])
        self.assertEqual(100, self.manager_cacheOn.get_variants("P42685")[0]["location"])
                
    def test_get_gene_name(self):
        self.assertEqual("vpu", self.manager_cacheOff.get_gene_name("P05923"))
        self.assertEqual("vpu", self.manager_cacheOn.get_gene_name("P05923"))

    def test_get_geneID(self):
        self.assertEqual("2444", self.manager_cacheOff.get_geneID("P42685"))
        self.assertEqual("2444", self.manager_cacheOn.get_geneID("P42685"))

    def test_get_host(self):
        self.assertEqual("Homo sapiens (Human)", self.manager_cacheOff.get_host("P05923"))
        self.assertEqual("Homo sapiens (Human)", self.manager_cacheOn.get_host("P05923"))


    
    def test_get_protein_sequence_length(self):
        self.assertEqual(505, self.manager_cacheOff.get_protein_sequence_length("P42685"))
        self.assertEqual(505, self.manager_cacheOn.get_protein_sequence_length("P42685"))
        
    def test_get_ID_type(self):
        self.assertEqual("UniProtKB/Swiss-Prot", self.manager_cacheOff.get_ID_type("P42685")[1])
        self.assertEqual("UniProtKB/Swiss-Prot", self.manager_cacheOn.get_ID_type("P42685")[1])

    def test_run_translation(self):
        self.assertEqual("1169745", self.manager_cacheOff.run_translation("P42685"))
        self.assertEqual("1169745", self.manager_cacheOn.run_translation("P42685"))

    def test_purge(self):
        self.manager_cacheOff.purge()
        self.manager_cacheOn.purge()
        self.assertEqual(0, self.manager_cacheOff.datastore_size())
        self.assertEqual(0, self.manager_cacheOn.datastore_size())

    def test_batch_get_variants(self):
        IDLIST = ["Q99439", "Q86UX7"]
        self.assertEqual(0, len(self.manager_cacheOff.batch_get_variants(IDLIST)["Q99439"]))
        self.assertEqual(0, len(self.manager_cacheOff.batch_get_variants(IDLIST)["Q86UX7"]))
        self.assertEqual(0, len(self.manager_cacheOn.batch_get_variants(IDLIST)["Q99439"]))
        self.assertEqual(0, len(self.manager_cacheOn.batch_get_variants(IDLIST)["Q86UX7"]))
    
    def test_batch_get_protein_name(self):
        IDLIST = ["Q08AM6", "P35582"]
        self.assertEqual(59, len(self.manager_cacheOff.batch_get_protein_name(IDLIST)["Q08AM6"]))
        self.assertEqual(105, len(self.manager_cacheOff.batch_get_protein_name(IDLIST)["P35582"]))
        self.assertEqual(59, len(self.manager_cacheOn.batch_get_protein_name(IDLIST)["Q08AM6"]))
        self.assertEqual(105, len(self.manager_cacheOn.batch_get_protein_name(IDLIST)["P35582"]))

    def test_Uniprot_fallback(self):
        IDLIST = ["P42858", "A6NC57"]
        self.assertEqual("Huntingtin; AltName: Full=Huntington disease protein; Short=HD protein", self.manager_cacheOn.batch_get_protein_name(IDLIST)["P42858"])
        self.assertEqual("Ankyrin repeat domain-containing protein 62", self.manager_cacheOn.batch_get_protein_name(IDLIST)["A6NC57"])
        

    def test_batch_get_protein_sequence(self):
        IDLIST = ["Q6P5D3", "Q14980"]
        self.assertEqual(1388, len(self.manager_cacheOff.batch_get_protein_sequence(IDLIST)["Q6P5D3"]))
        self.assertEqual(2115, len(self.manager_cacheOff.batch_get_protein_sequence(IDLIST)["Q14980"]))
        self.assertEqual(1388, len(self.manager_cacheOn.batch_get_protein_sequence(IDLIST)["Q6P5D3"]))
        self.assertEqual(2115, len(self.manager_cacheOn.batch_get_protein_sequence(IDLIST)["Q14980"]))


    def test_batch_get_edge_cases(self):
        IDLIST = []
        self.assertEqual({}, self.manager_cacheOn.batch_get_protein_sequence(IDLIST))
       
if __name__ == '__main__':
    print "Please run all tests together"



# Copyright 2012-2015 by Alex Holehouse - see LICENSE for more info
# Contact at alex.holehouse@wustl.edu

""" Acts as a easy-to-use wrapper around the Entrez package, for an intuitive
and simple way to get sequence information from NCBI. Abstracts NCBI's
eUtils from the user, providing a uniform interface. The absolute goal here is 
not to provide a customizable or complex system, but to provide a limited but
very easy to use small set of tools for getting the necessary data.

Variables:
email      Set the Entrez email parameter (default is none)
tool       Set the Entrez tool parameter (default is biopython)
version    Current version (string)
author     Author's email address

Internal Modules - call using Geeneus.Genome or Geeneus.Proteome
genome     Functions for dealing with genome data (e.g. getting sequences,
           coding regions)
proteome   Functions for dealing with protein data (e.g. protein sequence)

"""

import Proteome
import Genome

tool = "biopython"
email = None
version = "0.1.9"
author = "alex.holehouse@gmail.com"
print version



# Deals with cache and storing of protein data objects, as well as coordinating network access  (private)
#
# Copyright 2012-2015 by Alex Holehouse - see LICENSE for more info
# Contact at alex.holehouse@wustl.edu
import time
import sys
import re
import pickle

# Biopython selective imports
import Bio.Entrez.Parser
from Bio import Entrez
from Bio.Seq import Seq
from Bio.Alphabet import IUPAC
from xml.dom.minidom import parseString
import ProteinObject
import Networking
import httplib
import Parser as GRP
import UniprotAPI
from DataStructures import CaseInsensitiveDict as CID

IDTYPES = {-2:"International Protein Index",\
                -1:"Unknown protein accession type", \
                0:"GI", \
                1:"RefSeq", \
                2:"UniProtKB/Swiss-Prot", \
                3:"DDBJ", \
                4:"GenBank",\
                5:"EMBL",\
                6:"PDB",\
                7:"UniProtKB/Swiss-Prot",\
                8:"Unknown protein accession type",\
                }



######################################################### 
#########################################################
class ProteinRequestParser(GRP.GeneralRequestParser):

# Initialization function, set Entrez.email for calls, an ensure key -1 is set
# to a non-existant object
#
    def __init__(self, email, cache, retry=0, loud=True, shortcut=True):
        """"Initializes an empty requestParser object, setting the Entrez.email field and defining how many times network errors should be retried"""
        try:
            GRP.GeneralRequestParser.__init__(self, email, cache, retry, loud)
        
            self.protein_datastore = {-1 : ProteinObject.ProteinObject(-1, [])}
            self.protein_translationMap = {-1: --1}
            self.batchableFunctions = [self.get_sequence, self.get_protein_name, self.get_variants, self.get_geneID, self.get_protein_sequence_length]
            self.UniprotAPI = UniprotAPI.UniprotAPI()
            
            self.shortcut = shortcut
            self.error_status = False
             
        except Exception, e: 
            print "Fatal error when creating ProteinRequestParserObject"
            self.error_status = True
            raise e



#--------------------------------------------------------
# PUBLIC FUNCTION
#--------------------------------------------------------
# return a copy of the list of keys (without -1 key)
#
    def keys(self):
        kl = []
        for i in self.protein_datastore:
            kl.append(i)

        kl.remove(-1)
        return kl

#--------------------------------------------------------
# PUBLIC FUNCTION
#--------------------------------------------------------
# return a copy of the list of keys (without -1 key)
#
    def accession_classes(self):
        return IDTYPES.values()

#--------------------------------------------------------
# PUBLIC FUNCTION
#--------------------------------------------------------
# Test if the datatstore has a key (does not trigger download
# on fail

    def has_key(self, ID):
        return self.protein_datastore.has_key(ID)

            
#--------------------------------------------------------
# PUBLIC FUNCTION
#--------------------------------------------------------
#
# Check for an error in the parser object
    def error(self):
       return self.error_status


#--------------------------------------------------------
# PUBLIC FUNCTION
#--------------------------------------------------------
#
# Check for an error in the proteinObject obtained from
# the accession ID
    def protein_error(self, ID):
       ProtObj = self._get_protein_object(ID)
       return ProtObj.error()


#--------------------------------------------------------
# PUBLIC FUNCTION
#--------------------------------------------------------
#
# Check for an error in the proteinObject obtained from
# the accession ID
    def protein_exists(self, ID):
       ProtObj = self._get_protein_object(ID)
       return ProtObj.exists()

#--------------------------------------------------------
# PUBLIC FUNCTION
#--------------------------------------------------------
#
# Check for an error in the proteinObject obtained from
# the accession ID
    def get_creation_date(self, ID):
       ProtObj = self._get_protein_object(ID)
       return ProtObj.get_creation_date()


#--------------------------------------------------------
# PUBLIC FUNCTION
#--------------------------------------------------------
#
# Check for an error in the proteinObject obtained from
# the accession ID
    def get_protein_source(self, ID):
       ProtObj = self._get_protein_object(ID)
       return ProtObj.source()

#--------------------------------------------------------
# PUBLIC FUNCTION
#--------------------------------------------------------
#
# Get the protein's name
    def get_protein_name(self, ID):
        return (self._get_protein_object(ID)).get_protein_name()


#--------------------------------------------------------
# PUBLIC FUNCTION
#--------------------------------------------------------
#
# Get the protein sequence version
    def get_record_version(self, ID):
        return (self._get_protein_object(ID)).get_version()


#--------------------------------------------------------
# PUBLIC FUNCTION
#--------------------------------------------------------
#
# Get the AA sequence of the protein (N-to-C)
    def get_sequence(self, ID):
        ProtObj = self._get_protein_object(ID)
        return ProtObj.get_protein_sequence()
                
                       
#--------------------------------------------------------
# PUBLIC FUNCTION
#--------------------------------------------------------
#
# get a list of the single AA change varants for this
# species
    def get_variants(self, ID):
        ProtObj = self._get_protein_object(ID)
        return ProtObj.get_variants()


#--------------------------------------------------------
# PUBLIC FUNCTION
#--------------------------------------------------------
#
# get the GeneID associated with this protein
    def get_geneID(self, ID):
        ProtObj = self._get_protein_object(ID)
        return ProtObj.get_geneID()


#--------------------------------------------------------
# PUBLIC FUNCTION
#--------------------------------------------------------
#   
# Get the sequence length of the protein's AA sequence
    def get_protein_sequence_length(self, ID):
        ProtObj = self._get_protein_object(ID)
        return ProtObj.get_protein_sequence_length()


#--------------------------------------------------------
# PUBLIC FUNCTION
#--------------------------------------------------------
#   
# Get the sequence length of the protein's AA sequence
    def get_other_accessions(self, ID):
        ProtObj = self._get_protein_object(ID)
        return ProtObj.get_other_accessions()


#--------------------------------------------------------
# PUBLIC FUNCTION
#--------------------------------------------------------
#   
# Get the sequence length of the protein's AA sequence
    def get_species(self, ID):
        ProtObj = self._get_protein_object(ID)
        return ProtObj.get_species()

#--------------------------------------------------------
# PUBLIC FUNCTION
#--------------------------------------------------------
#   
# Get the sequence length of the protein's AA sequence
    def get_taxonomy(self, ID):
        ProtObj = self._get_protein_object(ID)
        return ProtObj.get_taxonomy()

#--------------------------------------------------------
# PUBLIC FUNCTION
#--------------------------------------------------------
#   
# Get the sequence length of the protein's AA sequence
    def get_host(self, ID):
        ProtObj = self._get_protein_object(ID)
        return ProtObj.get_host()

#--------------------------------------------------------
# PUBLIC FUNCTION
#--------------------------------------------------------
#   
# Get the sequence length of the protein's AA sequence
    def get_domains(self, ID):
        ProtObj = self._get_protein_object(ID)
        return ProtObj.get_domains()

#--------------------------------------------------------
# PUBLIC FUNCTION
#--------------------------------------------------------
#   
# Get the sequence length of the protein's AA sequence
    def get_gene_name(self, ID):
        ProtObj = self._get_protein_object(ID)
        return ProtObj.get_gene_name()

#--------------------------------------------------------
# PUBLIC FUNCTION
#--------------------------------------------------------
#
# Get any isoforms associated with this sequence ID

    def get_isoforms(self, ID):
        ProtObj = self._get_protein_object(ID)
        return ProtObj.get_isoforms()
#--------------------------------------------------------
# PUBLIC FUNCTION
#--------------------------------------------------------
#
# Function to translate and accession value into a GI. If 
# there is more than one possible GI, then simply returns
# -1. Caches lookup values so we don't have to repeatedly 
# run the translation remotely

    def translate_Asc2GI(self, Accession):
        
        retry = self._build_retry_function(self.Networking.esearchProtein)
      
        if not Accession in self.protein_translationMap:
            record = -1

            while (record == -1):
                record = retry(Accession)

            # if we retry a bunch of times and it doesn't work
            if record == -2:
                print "Unable to carry out esearch for term{y}".format(y=Accession)
                return -1
                        
            # else we got some XML
            IdList = record["IdList"]

            # if there's only one GI associated with this accession number then great!
            if len(IdList) == 1:
                self.protein_translationMap[Accession] = str(record["IdList"][0])

            # if not
            else:
                print "There are {op} possible options, shown below. For PDB values, this often arises because seperate chains are treated as different proteins".format(op=len(IdList))
                for i in IdList:
                    print i
                return -1

        return self.protein_translationMap[Accession]


#--------------------------------------------------------
# PUBLIC FUNCTION
#--------------------------------------------------------
#
# Function to get the raw XML 
    def get_raw_xml(self, ProteinID):

        ProtObj = self._get_protein_object(ProteinID)
        return ProtObj.get_raw_xml()


#--------------------------------------------------------
# PUBLIC FUNCTION
#--------------------------------------------------------
# Function to get the raw XML 
# 
    def get_ID_type(self, ProteinID):
        return ID_type(ProteinID)


#--------------------------------------------------------
# PUBLIC FUNCTION
#--------------------------------------------------------
# Function which purges the datastore
#
    def purge_data_store(self):
        del self.protein_datastore
        self.protein_datastore = {-1 : ProteinObject.ProteinObject(-1, [])}

#--------------------------------------------------------
# PUBLIC FUNCTION
#--------------------------------------------------------
# Get the number of items in the datastore
#
    def datastore_size(self):
        return len(self.protein_datastore)-1



#--------------------------------------------------------
# PUBLIC FUNCTION
#--------------------------------------------------------
# Save the datastore to file
#
    def save_datastore(self,filename='datastore'):

        with open(str(filename)+".pi",'w') as FH:
            pickle.dump(self.protein_datastore, FH)


#--------------------------------------------------------
# PUBLIC FUNCTION
#--------------------------------------------------------
# Load a datastore from file
#

    def load_datatore(self,filename):
        with open(str(filename),'r') as FH:
            self.protein_datastore = pickle.load(FH)








#--------------------------------------------------------
# PUBLIC FUNCTION
#--------------------------------------------------------
#

    def batchFetch(self, function, listOfIDs):
        outputDict = {}
        toFetch = []

        # check the function actually makes sense taking
        # a single ID element as input
        if function not in self.batchableFunctions:
            print "Warning, function cannot be run via batch"
            return

        # Firstly, we identify which, if any of these are already in the 
        # datastore, and for those which are not ignore badly formatted
        # accession numbers
        for proteinID in listOfIDs:
            if proteinID not in self.protein_datastore or not self.cache:
                proteinID = self._convertIfNecessary(proteinID)
                if not ID_type(proteinID)[0] < 0:
                    toFetch.append(proteinID)                    

        
        
        # if we're running using shortcutting we have to now split to toFetch list into a
        # shortcutable list and a non-shortcutable list, deal with each list seperatly, them
        # re-assemble them into the same order.
        if self.shortcut:
            trackerVector = []
            UniProtList = []
            NCBIList = []
            listOfXML = []
            
            # build two distinct lists and a tracking vector which describes the mapping
            # of toFetch to those two lists
            for ID in toFetch:
                if ID_type(ID)[0] == 7:
                    trackerVector.append("U")
                    UniProtList.append(ID)
                else:
                    trackerVector.append("N")
                    NCBIList.append(ID)

            # build a XMLlist using the NCBI batch lookup function
            NCBI_XML = self._get_batch_XML(NCBIList, self.Networking.efetchProtein, self.UniprotDatabaseLookup)
            
            # silently adds records to database
            self.UniprotAPI.batchFetch(UniProtList, self.protein_datastore)
            
            # now reconstruct listOfXML as a list the same order as toFetch, but with all the would-be UniProt
            # XML values equal to -1
            NCBI_counter = 0
            for i in trackerVector:
                if i == "U":
                    listOfXML.append(-1)
                if i == "N":
                    listOfXML.append(NCBI_XML[NCBI_counter])
                    NCBI_counter = NCBI_counter+1

        else:
                           
            # next we take those which are NOT in the datastore and take advantage
            # of the Biopython.Entrez' batch downloaded function. This makes a SINGLE
            # call to the server, so is a lot faster (reduces setup and teardown)
            # generates a list, each element of which is the 1:1 xml for the 
            # listOfIDs
            listOfXML = self._get_batch_XML(toFetch, self.Networking.efetchProtein, self.UniprotDatabaseLookup)
        
        
        # note we don't have to try/catch here because _get_batch_XML() guarentees
        # that each XML field is valid
        fetchCounter = 0
        
        for protein_xml in listOfXML:
            if not self.protein_datastore.has_key(toFetch[fetchCounter]) and not protein_xml == -1 :
                self.protein_datastore[toFetch[fetchCounter]] = ProteinObject.ProteinObject(toFetch[fetchCounter], [protein_xml])
            
            fetchCounter = fetchCounter+1

                        
        # finally, we build a tuple with results from everything in the input, and
        # return. Note we're ONLY doing this if that ID has already been loaded into
        # the data store. This stops the retry mechanism going through for batch, failing
        # and then just retrying again via the _get_protein_object() function which the
        # $function inevitably will call
        #
        # By checking against the datastore we also ensure we can pick up additional records
        # built by the "alternative" function, which adds to the datastore in an independent manner
        # to the for loop above 
            
        for ID in listOfIDs:
            if ID in self.protein_datastore:
                outputDict[ID] = function(ID)
                
            else:
                outputDict[ID] = function(-1)

                
        # build a case insensitive read only dictionary to output
        outFinal = CID(outputDict)

        return outFinal

#--------------------------------------------------------
# PRIVATE FUNCTION FUNCTION
#--------------------------------------------------------
#

    def _get_protein_object(self, proteinID):

        proteinID = self._convertIfNecessary(proteinID)

        # pre-check to see if the accession matches the predefined accession format
        # rules. If it doesn't then 
        if ID_type(proteinID)[0] < 0:
            if not proteinID == -1: 
                self.printWarning("\nWarning - The ID {ID} is an invalid accession number, and the database will not be queried".format(ID=proteinID))
            proteinID = -1

        else:
            # note _get_object is inherited from Parser
            self._get_object(proteinID, self.protein_datastore, self._protein_fetch_function, ProteinObject.ProteinObject, self.UniprotDatabaseLookup)
        
        return self.protein_datastore[proteinID]

#--------------------------------------------------------
# PRIVATE FUNCTION FUNCTION
#--------------------------------------------------------
# The protein fetch function allows an additional layer
# of logic when deciding how to deal with an accession value. 
#
#
    def _protein_fetch_function(self, accessionID):

               
        # if we have shortcut off then just use eFetch. If eFetch fails after a number
        # of retries we'll try UniProt if the accession is a UniProt one
        if self.shortcut == False:
            returnVal = self.Networking.efetchProtein(accessionID)
        
        # else we have shortcutting switched on, so if the accession is
        # a uniprot accession which NCBI doesn't guarentee to support we
        # automaically drop 
        else:
            if ID_type(accessionID)[0] == 7:

                
                # because we provide the getProteinObjectFromUniProt with a -1 value for 
                # the datastore, the function will return an instantiated ProteinObject
                # if possible (or -1 on failure)
                returnVal = self.UniprotAPI.getProteinObjectFromUniProt(-1, accessionID)

            else:
                returnVal = self.Networking.efetchProtein(accessionID)
               
        return returnVal
 
                


#--------------------------------------------------------
# PRIVATE FUNCTION
#--------------------------------------------------------          
# Function which we can manually add cases to where converting
# from some kind of accession number to a GI would be better
# for whatever reason. Currently only true for PDB IDs
#

    def _convertIfNecessary(self, ProteinID):
        
        # Convert to upper case for tests so we have
        # a homogenous string
        try:
            ProteinID = ProteinID.upper()
        except AttributeError:
            return ProteinID

        # if PDB translate to a GI value
        if ID_type(ProteinID)[0] == 6:
            return self.translate_Asc2GI(ProteinID)
        
        # if SwissProt/UniProt with an isoform identifier
        # chop off identifier
        if re.match("[A-Z][0-9][A-Z0-9][A-Z0-9][A-Z0-9][0-9]-[0-9]", ProteinID):
            return ProteinID[:6]
        
        return ProteinID

#--------------------------------------------------------
# PRIVATE FUNCTION
#--------------------------------------------------------
# Function to print a message if loud has been set to 
# true - saves littering code with if self.loud: statements
#
    def printWarning(self, message):
        if self.loud:
            print message


#--------------------------------------------------------
# PRIVATE FUNCTION
#--------------------------------------------------------
# Function which gives the GeneralRequestParser access to
# non NCBI based means to get a protein record (Uniprot)
#

    def UniprotDatabaseLookup(self, accessionID):
        
        IDtype = ID_type(accessionID)[0]
        
        # if shortcut is on and we're calling the alt function
        # we don't want to try UniProt *again* because it must
        # have not worked the first few times
        if self.shortcut:
            exitcodes = 2,

        # however if shortcut is off then we do want to test any
        # UniprotKB/Swissprot accession
        else:
            exitcodes = 2,7
        
        if IDtype in exitcodes:
            print "[UniProt]: Falling back and querying UniProt servers...  "
        
            # query the server through the UniprotAPI class
            self.UniprotAPI.getProteinObjectFromUniProt(self.protein_datastore, accessionID)
                    
            # NOTE this is crucial - we only return true if we can get an object 
            # from the dictionary
            try:
                self.protein_datastore[accessionID]
                print "[UniProt]: Sucess!"
                return True
            except KeyError:
                return False
        else:
            return False


    def loadfile(self, filename):
        with open(filename) as openFile:
            print "Reading in file..."
            xmlstring = openFile.read()
           

            compositeDOM = parseString(xmlstring)
           
        print "Done"
        print "Extracting protein data..."
        elementsList = compositeDOM.getElementsByTagName("entry")
        
        
        total = len(elementsList)
        counter = 0 
        
        "Parsing XML..."

        for element in elementsList:
            print "Parsing " + str(counter) +"/"+str(total)
            self.UniprotAPI.sideload_from_file(self.protein_datastore, element)
            counter = counter+1
                
        

# +-------------------------------------------------------+
# |                    END OF CLASS                       |
# +-------------------------------------------------------+


#--------------------------------------------------------
# PRIVATE FUNCTION
#--------------------------------------------------------
## ProteinParser function
## General function true to the module, as does not require
## a class instance to work
#
# List of for refrence/addition
# -2 : IPI - IPI is -2 because we can't look it up, but it 
#            is a valid type of accession. Neither NCBI nor
#            Uniprot support IPI lookups, however!
# -1 : unknown
#  0 : GI
#  1 : refseq
#  2 : Swissprot
#  3 : DDBJ
#  4 : genbank
#  5 : EMBL
#  6 : PDB
#  7 : Uniprot


def ID_type(ProteinID):
    
    # convert ID to all uppercase characters, and remove any periods 
    try:
        ProteinID = ProteinID.upper()
    except AttributeError:
        return ([-1, IDTYPES[-1]])


    # cut off . and anything after
    # "." represent versions of accession numbers, but in this context 
    # the versions a) don't change the type and b) just make all the
    # regexes way harder!

    if ProteinID.find(".") > 0:
        ProteinID = ProteinID[:ProteinID.find(".")]

        
    # if it begin [A|N|X|Y|Z]P_ then it's a refseq 
    if re.match("[ANXYZ][P]_", ProteinID):
        return [1, IDTYPES[1]]

    # if the ID is all digits it's a GI
    if ProteinID.isdigit() or re.match("GI",ProteinID):
        return [0, IDTYPES[0]]

    # is it a PDB?
    if re.match("[0-9][A-Z0-9][A-Z0-9][A-Z0-9]", ProteinID) or re.match("[0-9][A-Z0-9][A-Z0-9][A-Z0-9]_[A-Z0-9]", ProteinID):
        return[6, IDTYPES[6]]
      
    # if it begins [O|P|Q] then it is found in NCBI database
    if re.match("[OPQ]", ProteinID) and re.match("^[A-Z0-9]+$", ProteinID) and len(ProteinID) == 6:
        return [2, IDTYPES[2]]

    # else if its a uniprot/swissprot is is not necessarily found so we fall back to UniProt
    # api
    if re.match("[A-N|R-Z][0-9][A-Z0-9][A-Z0-9][A-Z0-9][0-9]", ProteinID) and len(ProteinID) == 6:
        return [7, IDTYPES[7]]

    if re.match("IPI[0-9]*", ProteinID):
        return [-2, IDTYPES[-2]]
    
    # Now we've removed refseq, gi and swissprot we can be a bit more discerening about 
    # what the accession should be (as now it must conform to the NCBI accesison number
    # rules). If any of the following fail then we assume the accession is malformed
    # __________________________________
    # Should be 3 letters+5digits
    # Should only contain A-Z
    # Should be less than 8 characters
    # 
    if not re.match("[A-Z][A-Z][A-Z][0-9][0-9][0-9][0-9][0-9]", ProteinID) or \
            not re.match("^[a-zA-Z0-9]+$", ProteinID) or \
            len(ProteinID) > 8:
        return ([-1, "Unknown protein accession type"])
    
    # DDBJ
    if re.match("[B][A-Z][A-Z]", ProteinID) or re.match("F[A-Z][A-Z]", ProteinID) or \
            re.match("G[A-Z][A-Z]", ProteinID) or re.match("I[A-Z][A-Z]", ProteinID):
        return [3, IDTYPES[3]]    

    # GenBank
    if re.match("A[A-Z][A-Z]", ProteinID) or re.match("AAE", ProteinID) or \
        re.match("D[A-Z][A-Z]", ProteinID) or re.match("E[A-Z][A-Z]", ProteinID) or \
        re.match("H[A-Z][A-Z]", ProteinID) or re.match("J[A-Z][A-Z]", ProteinID):  
        return [4, IDTYPES[4]]

    if re.match("C[A-Z][A-Z]", ProteinID):
        return [5, IDTYPES[5]]
    
    # if we get here something seems to have gone wrong...
    return [-1, IDTYPES[-1]]
    

    



# Copyright 2012 - 2015 by Alex Holehouse - see LICENSE for more info
# Contact at alex.holehouse@wustl.edu

import collections

class CaseInsensitiveDict(collections.Mapping):
    def __init__(self, d):
        self._d = d
        self._s = dict((k.upper(), k) for k in d)
        self._lc = dict((k.upper(), d[k]) for k in d)
 
    def __contains__(self, k):
        return k.upper() in self._s
    
    def __len__(self):
        return len(self._s)
    
    def __iter__(self):
        return iter(self._s)
    
    def __getitem__(self, k):
        return self._d[self._s[k.upper()]]
    
    def actual_key_case(self, k):
        return self._s.get(k.upper())

    def __repr__(self): 
        return str(self._lc)



# Provides a bunch of general purpose functions for boring sys-related
# stuff
# Part of the Geeneus package
#
# Copyright 2012-2015 by Alex Holehouse, with significant contributions
# from Matt Matlock - see LICENSE for more info
# Contact at alex.holehouse@wustl.edu


def show_warning(msg):
    print "[WARNING]: " + str(msg)

def show_error(msg):
    print "[ERROR]: " + str(msg)

def show_status(msg):
    print "[STATUS]: " + str(msg)
    



import unittest
import time
from geeneus.backend import ProteinObject
from geeneus.backend import Networking
from geeneus.backend import UniprotAPI
from Bio import Entrez
import urllib2
from xml.dom.minidom import parseString

import random

class TestIsoformAlgorithm(unittest.TestCase):

    def setUp(self):

        self.nw = Networking.Networking(30)
        self.UP = UniprotAPI.UniprotAPI()
        Entrez.email = "alex.holehouse@gmail.com"
        
    def test_barrage_1(self):



        # -------------------------------------------------------------------------------
        # patho represents a set of IDs which caused failures for various reasons
        # over the system's development. They provide an initial previously pathalogical
        # dataset to test on before we hit random pulldowns.
        #
        # - P11362    really complicated, more algorithm correctness proof really 
        #             (seriously, even I'm amazed this works!)
        #
        # - P19020    closing bracket in name. Edgecase on naming conventions
        #
        # - Q13635    Has a "'" in name. Edgecase on naming conventions
        #
        # - Q14814    Also has "'" in name.
        #
        # - Q9QXS1    Straw that broke the camel's back - isoform names are full
        #             of commas. Forced a complete re-write of the name identification
        #             region, but hopefully the new code is totally robust and a *lot*
        #             easier to understand. Seriously, isoform 'PLEC-0,1C,2A,3A' - this
        #             is not an OK name!!
        #
        #
        # Other known pathalogical inputs
        # - Q9NY33    NCBI failed to get isoform '3'. Annoyingly this is because
        #             Uniprot has a '3' isoform, but NCBI doesn't. Not a lot we can
        #             do about that, really...
        # - O95467    NCBI has no splicing variant data at all, despite talking about
        #             it :-(
        #
        # - O14772    NCBI has no splicing variant data at all
        #
        # - Q14994    NCBI is missing any reference to isoform 8 (16th Jan 2013)
        #
        # ------------------------------------------------------------------------------
        patho = ["Q8IYH5", "Q9P0K8", "Q9UK53", "P60411", "Q9NP78", "P11362", "Q9NY33", "P19020", "Q13635", "Q9QXS1", "Q14814"]
        autofail = ["Q9NY33", "O95467", "O14772", "Q14994"]
        # 
        
        pathocounter = 0

        lengthOfTest = len(patho)+10
        
        for i in xrange(0,lengthOfTest):
            
            print "On " + str(i) + " of " + str(lengthOfTest)
            print "-----------------------------------------------------------------------------------------"

            if pathocounter < len(patho):
                ID = patho[pathocounter]
                pathocounter = pathocounter+1
            else:
                ID = self.get_random_accession()
           
            print ID
            
            # Some values are just unsolvably pathalogical, so we allow the test to skip them. By unsolvable,
            # this is where a record is just missing from one of the databases - basically a human error or
            # some kind of inconsistency means we'll never get the same isoforms.
            #
            if ID in autofail:
                print "Database error (i.e. there's some reason this will *never* work - skipping"
                continue
            
            NCBIISO = self.get_NCBI_isoform(ID)
            if NCBIISO == -1:
                # If we have a problem looking up a UniProt record
                # just skip it
                continue

            print "NCBI lookup done..."

            UNIISO = self.get_Uniprot_isoform(ID)
            if UNIISO == -1:
                # If we have a problem looking up a UniProt record
                # just skip it
                continue

                        
            print "UniProt lookup done"
            
            ## useful for failure analysis
           # print "NCBI dictionary"
           # print NCBIISO
           # print "UniProt dictionary"
           # print UNIISO
            
            self.assertEqual(self.heuristic_compare(NCBIISO, UNIISO), True)
            #print "\=====Isoform sequences match!===="

    def get_random_accession(self):
        # 27 000 000 IDs, so we randomly select 1 between 1 and 20 000 000
        go = False

        while go == False:
            try:
                query = "http://www.uniprot.org/uniprot/?query=reviewed:yes+AND+organism:9606&format=xml&limit=1&offset="+str(random.randint(1,20000))+"&random=yes"
                handle = urllib2.urlopen(query)
                dom = parseString(handle.read())
                go = True
            except:
                continue
            
        return str(dom.getElementsByTagName("accession")[0].childNodes[0].nodeValue)


    def get_NCBI_isoform(self, ID):
        handle = self.nw.efetchProtein(ID)
        if not handle == -1:
            XML = Entrez.read(handle)
            PO = ProteinObject.ProteinObject(ID, XML)
            return(PO.get_isoforms())
        else:
            return -1

    def get_Uniprot_isoform(self, ID):
        datastore = {}
        PO = self.UP.getProteinObjectFromUniProt(datastore, ID)
        
        if datastore.has_key(ID):
            return(datastore[ID].get_isoforms())
        else:
            return -1
        

    def heuristic_compare(self, NCBI, UNIPROT):
        
        useRebased = False;

        if not len(NCBI) == len(UNIPROT):
            return False

        for k in UNIPROT.keys():
            try:
                seq1 = UNIPROT[k][1]
                seq2 = NCBI[k][1]
                
            except KeyError:
                print "Key mismatch"
                print "Uniprot keys are = " + str(UNIPROT.keys())
                print "NCBI key are = " + str(NCBI.keys())
                return False
                
            if not seq1 == seq2:
                return False

        return True
        



# Base class which deals with caching and storing of data objects, as well as coordinating network access (private)
#
# Copyright 2012-2015 by Alex Holehouse - see LICENSE for more info
# Contact at alex.holehouse@wustl.edu

from Bio import Entrez
from Bio.Seq import Seq
from Bio.Alphabet import IUPAC

import httplib

import Networking


## This exception is imported by other classes, and is used
## when they are creating new objects and discover, to their
## horror, that the XML is semantically wrong (even if it's
## syntactically good) - e.g. if it's missing some expected
## info
class BadXMLException(Exception):
    pass


class GeneralRequestParser:
    
    def __init__(self, email, cache, retry=0, loud=True):
        try:
            Entrez.email = email
            self.Networking = Networking.Networking(30)
            self.loud = loud
            self.retry = retry
            self.cache = cache
            self.error_status = False
        
        except: 
            print "Error building generic parser object"
            self.error_status = True
    

            ###
    def error(self):
        return self.error_status
    
    
    #--------------------------------------------------------
    # Preconditions
    # ID must be a valid ID, so probably want to run through a preprocessing function before you
    # call this
    #
    # - datastore is the datastore passed from the child class, which must be an ID indexed dictionary
    #   where the value is a specific storage object type (e.g. ProteinObject, GeneObject etc)
    #
    # - fetchFunction is the Networking function which we call to
    #
    # - newObjectConstructor is a constructor for the object type we store in datastore
    #
    # - alternative is the "alternative function" which we query if NCBI services don't seem to work
    # 
    def _get_object(self, ID, datastore, fetchFunction, newObjectConstructor, alternative=False):
        
        altSuccess = False  # set flag for later
      
        if ID == -1 or ID not in datastore or not self.cache:  
           
            if ID  == -1:
                
                # by returning the object associated with [] we don't pollute the datastore with invalid and pointless
                # searches, we avoid queriying NCBI without a hope in hell of a hit, and we take advantage of the built in
                # bad XML behaviour without raising an error, because, technically, no error has happened, we just know
                # that the ID in question won't return valid XML. It's not an error - it's just stupid.
                
                return newObjectConstructor(-1, [])
                       
            xml = -1
            retry = self._build_retry_function(fetchFunction);
                     
            # retry uses a closure based approach, so this just happens a set number of times
            # - not an infinite loop if we can't get a response!
            while (xml == -1):
                xml = retry(ID)

            # if we still can't get through after retrying a number of times
            if (xml == -2):

                # If we have an alternative function to try let's try it!
                # Alternative functions have to be totally defined and initialized
                # in the calling child class. The function can, obviously, do whatever
                # you want, but the behaviour must generally;
                #
                # 1) Take just an accession ID (a datastore object is internally stored
                #    as a object variable from the child class)
                #
                # 2) Deal with everything else internally
                #
                # 3) Simply return a "True" if it was succesful, or "False" if not
                #
                # If alternative returns success it *MUST* have been able to query
                # the datastore with ID and return a relevant object.
                #
                
                if alternative:
                    altSuccess = alternative(ID)

                if not altSuccess:
                    datastore[ID] = newObjectConstructor(-1, [])
                
            # or NCBI query was succesfull
            else:
                # if the xml has been set to an object, this object MUST be
                # the already constructed *Object (e.g. ProteinObject, GeneObject etc)
                # allows additonal control in the fetch function
                if str(type(xml)) == "<type 'instance'>":
                    datastore[ID] = xml

                # else parse that XML
                else:

                    # this try catches cases where the XML, while formatted correctly,
                    # actually doesn't make sense or is lacking in an expected field
                    try:
                        datastore[ID] = newObjectConstructor(ID, xml)
                    except BadXMLException:
                        datastore[ID] = newObjectConstructor(-1, [])
                
        return datastore[ID]

    
#--------------------------------------------------------
# PRIVATE FUNCTION
#-------------------------------------------------------
# This function provides an interface with the Networking
# functionality to return a list of XML elements, each of
# which corresponds to the ID in the listofIDs
#
#
    def _get_batch_XML(self, listOfIDs, function_to_apply, alternative=False):
        
        # base case 1 (base case 2 if len == 1)
        if len(listOfIDs) == 0:
            return []

        xml = -1
        retry = self._build_retry_function(function_to_apply);

        while (xml == -1):
            xml = retry(listOfIDs)

        # If we fail implement recursive batch fetch, such that we split the list in half
        # and batch fetch each half. Do this after troubleshooting!
        if xml == -2 or not len(xml) == len(listOfIDs):

            # we only call alternative if we're down to a single accession - ie. really
            # try and get it from NCBI
            if len(listOfIDs) == 1:
                if alternative:
                    alternative(listOfIDs[0])
                # note we still return -1 so as to allow the algorithm to be maximally efficient
                return [-1]
            
            # split list in half and recursivly batch both halves
            # this serves two purposes - it lets us retry a number of times proportional
            # to the length of the list, and it provides a binary search mechanism to 
            # root out a potential rotten apple which may be causing the list to error

            xml = self._get_batch_XML(listOfIDs[:int(len(listOfIDs)/2)], function_to_apply, alternative)
            xml_2 = self._get_batch_XML(listOfIDs[int(len(listOfIDs)/2):], function_to_apply, alternative)

            xml.extend(xml_2)
            
        return xml


#--------------------------------------------------------
# PRIVATE FUNCTION
#--------------------------------------------------------          
# Builds a closure based retry function, which when called
# the first self.retry times will atempt to get the parsed
# xml for the ID in question. However, on the
# self.retry+1 time it will simply return -2 
#
#
# Return Values
#     -1 on failure
#     An initialized GeeneusObject
#     Entrez XML structure
#
    def _build_retry_function(self, function_to_apply):

        retryCounter = [0]
        numberOfRetries = self.retry

        def retry(ID):

            if retryCounter[0] < numberOfRetries+1:
                
                retryCounter[0] = retryCounter[0]+1
                
                ## if we're not on our first try
                if not retryCounter[0]-1 == 0:
                    print("Retry number " + str(retryCounter[0]) + " of " + str(numberOfRetries+1))
                
                ## return value may be a real handle or -1
                handle = function_to_apply(ID)
                
                ## if we failed return -1
                if handle == -1:
                    return -1

                # function to apply can be made to return either XML
                # or a fully formed GeeneusObject (e.g. shortcutting
                # for ProteinObject).
                try:
                    if handle.type == 'GeeneusObject':
                        return handle
                except AttributeError:
                    pass # carry on...
                
                try:
                    XML = Entrez.read(handle)
                
                # what are these errors?
                # httplib.IncompleteRead - we accidentally closed the session early, either because of a timeout on the client end (i.e. here) or because 
                #                          of some kind of server error
                # 
                # Bio.Entrez.Parser.CorruptedXMLError - Something is wrong with the XML 
                # Bio.Entrez.Parser.NotXMLError - the XML is not XML (unlikely, but worth keeping!)
                # Bio.Entrez.Parser.ValidationError - unable to validate the XML (this can be ignored, but best not to!)
                except (httplib.IncompleteRead, Entrez.Parser.CorruptedXMLError, Entrez.Parser.NotXMLError, Entrez.Parser.ValidationError), err:  
                    return -1

                return XML

            else:
                return -2;

        return retry        



# UniProt server access layer. Carries out all DOM compliant
# XML processing returned by the UniProt networking requests
#
# Copyright 2012 - 2015 by Alex Holehouse - see LICENSE for more info
# Contact at alex.holehouse@wustl.edu
#

import urllib2
from xml.dom.minidom import parseString
import ProteinObject
import Networking
import ProteinParser
import StringIO
from Bio import SeqIO
import re



######################################################### 
######################################################### 
# Exception class for UniProt record access
#
class UniprotAPIException(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)    

######################################################### 
######################################################### 
# Main class for UniProt record parsing
#
class UniprotAPI:

    def __init__(self):   
        self.Network = Networking.Networking(40)
    

#--------------------------------------------------------
# PUBLIC FUNCTION
#--------------------------------------------------------
# Main public function for this class, and the one called by
# the ProteinParser object. Takes an existing datastore (i.e.
# a dictionary) and an accession ID, and using that accession
# ID queries the UniProt servers, if it can obtain the relevant
# record updates the $dataStore with the data, and if not does not
# 
# This function's return value/behaviour depends on the value of
# datastore. If datastore == -1 then it is assumed the function
# needs to return the initialized object. If not, then it is
# assumed the function is to update the datastore and give
# no return value.
#
# Note the majority of the coding logic is in _build_and_complete
# but we've seperated this into two functions so we can feed
# _build_and_code raw DOM objects obtained through the batch method
#
    def getProteinObjectFromUniProt(self, datastore, accessionID):

        # networking and deal with UniProtNetworkRequest XML
        # dom should be a DOM compliant XML object, produced 
        # by the minidom XML parser
        dom = self.getDOMObject(accessionID)


        return self._build_and_complete(accessionID, dom, datastore)

    
    def sideload_from_file(self, datastore, dom):
        
        childNodes = dom.childNodes
        
        accession = ""

        for i in childNodes:
            try:
                if str(i.nodeName) == "accession":
                    accession = str(i.childNodes[0].nodeValue).upper()
                    break
            except:
                continue
            
        if accession == "":
            print "Unable to find accession tag"
            return False

        return(self._build_and_complete(accession, dom, datastore))

                

    
    def _build_and_complete(self, accessionID, dom, datastore):
        
        # set return value if an error occurs
        if datastore == -1:
            failure = -1
        else:
            failure = None
            
        try:
            xml = dom.toxml()
        except AttributeError:
            print "Error - unable to parse xml associated with " + accessionID
            return failure

        # If we get here then the networking portion theoretically worked
        # Next try and parse the XML for the various elements needed for the ProteinObject explicit
        # initializer
        try:
            name = self._getProteinName(dom, accessionID)
            geneID = self._getProteinGeneID(dom, accessionID)
            geneName = self._getProteinGeneName(dom, accessionID)
            sequence = self._getProteinSequence(dom, accessionID)
            version = self._getVersion(dom, accessionID)
            mutations = self._getProteinMutations(dom, sequence)
            creationDate = self._getCreationDate(dom, accessionID)
            other_accessions = self._getOtherAccessionValues(dom, accessionID)
            species = self._getSpecies(dom, accessionID)
            domains = self._getDomains(sequence, accessionID)
            host = self._getHost(dom, accessionID)
            taxonomy = self._getTaxonomy(dom, accessionID)
            isoforms = self._getIsoforms(dom, accessionID)
        except UniprotAPIException, e:
            print e
            print "Error while parsing UniProt XML associated with accession + " + accessionID + ".\nSkipping..."
            
            # this means if we error at any point we don't add it to the datastore, but if all is well we do
            return failure
        
        # so, if we got here were able to parse the XML ok, so jobs a good'un!
        # return object
        if datastore == -1:
            return ProteinObject.ProteinObject(accessionID, version, xml, name, mutations, sequence, creationDate, geneID, geneName, other_accessions, species, domains, taxonomy, isoforms, "UniProt", host)

        # update datatstore and return silently
        else:
            datastore[accessionID] = ProteinObject.ProteinObject(accessionID, version, xml, name, mutations, sequence, creationDate, geneID, geneName, other_accessions, species, domains, taxonomy, isoforms, "UniProt", host)


#--------------------------------------------------------
# PUBLIC FUNCTION
#--------------------------------------------------------
# Public facing batch access function - updates the datastore
#

    def batchFetch(self, IDLIST, datastore):
        
        entryList = self._internal_batch_fetch(IDLIST)
        
        counter = 0
        for dom in entryList:
            self._build_and_complete(IDLIST[counter], dom, datastore)
            counter = counter+1


    def _internal_batch_fetch(self, IDLIST):
        if len(IDLIST) == 0:
            return []

        xmlstring = self.Network.UniProtBatchIsoformNetworkRequest(IDLIST, 'xml')

        try:
            compositeDOM = parseString(xmlstring)
        except:
            if len(IDLIST) == 1:
                return [-1]

            split = len(IDLIST)/2
            
            L1 = self._internal_batch_fetch(IDLIST[:split])
            L2 = self._internal_batch_fetch(IDLIST[split:])
            return L1+L2

        return compositeDOM.getElementsByTagName("entry")



#--------------------------------------------------------
# PRIVATE FUNCTION
#--------------------------------------------------------
# Function which coordinates Networking. Takes an accession
# value, retries the networking on failure, and returns
# a DOM compliant DOM-XML object
#   
    def getDOMObject(self, accessionID):

        handle = -1
        counter=0
        
        # crude 3-times retry. Maybe build a ProteinParser like
        # retry method, but given UniProt isn't a primary search
        # database seems like overkill at the moment
        while(handle == -1):
            handle = self.Network.UniProtNetworkRequest(accessionID)
            
            # we get back not -1 try parsing the handle. This can fail 
            # in a whole bunch of ways, so if it does we simply catch
            # the exception and retry. Typically this fails if we get a
            # REALLY unlucky error, where *none* of the networking
            # systems catch it but the returned data is actually 
            # corrupt.
            if not handle == -1: 
                try:
                    domObject = parseString(handle.read())
                except Exception:
                    if counter==3:
                        return -1
                    handle = -1
                    counter = counter+1
                    continue
                
                # if all's well test and return the "entry" element
                try:
                    elements = domObject.getElementsByTagName("entry")
                    if len(elements) > 1:
                        print "[UniProt] Lookup for accession " + str(accessionID) + " generated more than one result. Skipping..."
                        return -1
                    retval = elements[0]
                    return retval
                except:
                    return -1
                    
            counter=counter+1
            if counter==3:
                return -1
    
    
#--------------------------------------------------------
# PRIVATE FUNCTION
#--------------------------------------------------------
# Internal function which gets the name of our protein from
# the DOM object
#
    def _getProteinName(self, domObject, ID):
        protein = domObject.getElementsByTagName('protein')
        self._neq1(protein, 'protein', ID)
        
        # Firstly we try and find names defined as recommendedName->fullName
        # 
        for childNode in protein[0].childNodes:
            if childNode.nodeName == "recommendedName":
                fullName = childNode.getElementsByTagName('fullName')
            
                # check we only have a single fullName name
                self._neq1_explicit(fullName, "Multiple 'fullName' tags under the 'recommendedName' tag found in UniProt XML file associated with accession number " + ID)
                return(fullName[0].firstChild.nodeValue)        
          
        # if we get here we couldn't find a clear winner using recommendedName we try submittedName
        #
        for childNode in protein[0].childNodes:
            if childNode.nodeName == "submittedName":
                fullName = childNode.getElementsByTagName('fullName')
            
                self._neq1_explicit(fullName, "Multiple 'fullName' tags under the 'recommendedName' tag found in UniProt XML file associated with accession number " + ID)

                return(fullName[0].firstChild.nodeValue) 
        
        raise UniprotAPIException("Unable to find a protein name in the XML associated with accession " + ID)
        



#--------------------------------------------------------
# PRIVATE FUNCTION
#--------------------------------------------------------
# Internal function which gets the protein sequence
# 
# Behaviour Notes
# - We select the "protein sequence" by choosing the sequence
#   tag which has a 'length' attribute. As far as I can tell the only
#   sequence tags with 'length' attributes are those which represent the
#   protein sequence, BUT this does sort of seem like a bit of a fragile 
#   way of doing things... This is why the $found variable is included.
#   It will trip and error if we have more than one sequence tag with a 
#   length element. 
#
#
#
    def _getProteinSequence(self, domObject, ID):
        found = False
        sequencelist = domObject.getElementsByTagName('sequence')

        for sequence in sequencelist:
            if sequence.hasAttribute("length"):
                if found:
                    raise UniprotAPIException("Found multiple 'sequence' tags which have a 'length' attribute. Means we are unable to select a unique sequence. This is a flaw in the XML parser - please subit a bug report!" + ID)
                    
                retval = str(sequence.firstChild.toxml()).translate(None, '\n')
                found = True

        return retval.lower()

#--------------------------------------------------------
# PRIVATE FUNCTION
#--------------------------------------------------------
# Internal function which gets the protein version

    def _getVersion(self, domObject, ID):
        found = False
        sequencelist = domObject.getElementsByTagName('sequence')

        for sequence in sequencelist:
            if sequence.hasAttribute("length"):
                if found:
                    raise UniprotAPIException("Found multiple 'sequence' tags which have a 'length' attribute. Means we are unable to select a unique sequence. This is a flaw in the XML parser - please subit a bug report!" + ID)
                    
                retval = str(sequence.attributes["version"].value)
                found = True

        return retval
                
#--------------------------------------------------------
# PRIVATE FUNCTION
#--------------------------------------------------------
# Internal function which gets the a list of dictionaries of protein mutations
# congruent with the ProteinObject variants list format
#

    def _getProteinMutations(self, domObject, sequence):
        mutations = []

        featureList = domObject.getElementsByTagName('feature')

        for feature in featureList:
            
            if feature.attributes['type'].nodeValue == 'sequence variant':

                tempDict = {}
                
                # first look for location->position nodes
                try:
                    tempDict['location'] = int(str(feature.getElementsByTagName('location')[0].getElementsByTagName('position')[0].attributes["position"].nodeValue))
                except IndexError, e:

                    # if that fails look for location->begin position nodes
                    try:
                        tempDict['location'] = int(str(feature.getElementsByTagName('location')[0].getElementsByTagName('begin')[0].attributes["position"].nodeValue))

                    # if *that* fails abort and raise an exception!
                    except IndexError, e:     
                        raise UniprotAPIException("Unable to find location list for mutation parsing")

                # assuming its not a deletion
                try:
                    tempDict['mutant'] = str(feature.getElementsByTagName('variation')[0].firstChild.toxml())
                    tempDict['original'] = str(feature.getElementsByTagName('original')[0].firstChild.toxml())
                    tempDict['variant'] = tempDict['original'] + " -> " + tempDict['mutant']
                    
                # for deletions
                except IndexError:
                    tempDict['mutant'] = "-"
                    tempDict['original'] = sequence[tempDict['location']-1].upper()
                    tempDict['variant'] = ""

                # The section below constructs a composite annotation field form
                # the description and the seperate mutation ID
                try:
                    desc = str(feature.attributes['description'].nodeValue)    
                except KeyError:
                    desc = ""

                try:
                    idDesc = str(feature.attributes['id'].nodeValue)
                    idDesc = " id="+idDesc
                except KeyError:
                    idDesc = ""

                tempDict['notes'] = desc + idDesc

                # Finally we determine the type of mutation

                if tempDict["mutant"] == "-":
                    tempDict["type"] = "Deletion"

                elif len(tempDict['mutant']) > len(tempDict['original']):
                    if tempDict['original'] in tempDict['mutant']:
                        tempDict["type"] = "Insertion"
                    else:
                        tempDict["type"] = "Insertion and substitution"

                elif len(tempDict['mutant']) < len(tempDict['original']):
                    tempDict["type"] = "Deletion & substitution"

                elif len(tempDict['mutant']) == len(tempDict['original']):
                    if len(tempDict['mutant']) == 1:
                        tempDict["type"] = "Substitution (single)"
                    elif len(tempDict['mutant']) == 2:
                        tempDict["type"] = "Substitution (double)"
                    else:
                        tempDict["type"] = "Substitution (" + str(len(tempDict['mutant'])) + ")"
                        
                # add this mutation to the list and continue
                mutations.append(tempDict)
                                              
        return mutations
                
                
#--------------------------------------------------------
# PRIVATE FUNCTION
#--------------------------------------------------------
# Internal function which gets the name of the gene associated with this protein
#
# Behaviour notes
# - Stop searching for genes after the first primary gene is found (i.e. ignore later primary genes. 
#   May want to add warning if multiple primary genes are found, although this shouldn't happen.
#
# - If no primary genes are found we automatically select the first gene element ($fallBack) as the
#   gene of interest
#
# - The gene name may not be known/recorded
##
    def _getProteinGeneName(self, domObject, ID):

        fallBack = ""
    
        geneList = domObject.getElementsByTagName('gene')
        
        # if there is no "gene" tag then set Gene name to an empty string
        try:
            self._neq1(geneList, 'gene', ID)
        except UniprotAPIException,e :
            print e
            return fallBack
        
        
        for subElement in geneList[0].getElementsByTagName('name'):
            
            # if fallBack is unset, set it (means fallBack is always set to the first real gene)
            if fallBack == "":
                fallBack = subElement.firstChild.toxml()
                
            # if we find the primary gene
            if subElement.attributes['type'].nodeValue == 'primary':
                return (subElement.firstChild.toxml())

        # if we get here no primary gene was found, so default to the first gene. Make sure we actually found 
        # a name somewhere in all that XML though!
        self._isEmpty(fallBack, "Unable to find a name tag associated with the UniProt XML for accession " + ID)
                
        return(fallBack)


#--------------------------------------------------------
# PRIVATE FUNCTION
#--------------------------------------------------------
# Internal function which gets the ID of the gene associated
# with this accession value
#
# Behaviour notes
# - Stop searching for genes after the first GeneID is found 
#   (there should only be one anyway)
#
# - If no IDs are found we just return an empty string
#
# - Gene ID may not be known/recorded
##

    def _getProteinGeneID(self, domObject, ID):
        elementList = domObject.childNodes
        for element in elementList:
            if element.nodeType == element.ELEMENT_NODE and element.getAttribute("type") == "GeneID":
                return element.getAttribute("id") 
        return ""

    


#--------------------------------------------------------
# PRIVATE FUNCTION
#--------------------------------------------------------
# Internal function which gets the date the record was created
#

    def _getCreationDate(self, domObject, ID):
        retval = ""

        retval = domObject.attributes['created'].nodeValue
        
        # raise an exception if we didn't find a creation date
        self._isEmpty(retval, "Unable to find a creation date in the UniProt XML string associated with accession number " + ID)
        
        return retval

#--------------------------------------------------------
# PRIVATE FUNCTION
#--------------------------------------------------------
# Internal function which gets a list of alternative accession 
# values associated with this record
#

    def _getOtherAccessionValues(self, domObject, ID):
        
        accList = []

        elementList = domObject.childNodes
        for element in elementList:
            
            # ________________________________________________________________
            # look for root XML based "accession" values
            if element.nodeName == 'accession':
                tempID = element.childNodes[0].nodeValue
                accList.append((str(tempID), ProteinParser.ID_type(tempID)[1]))

            # ________________________________________________________________
            # also look for a 'name' XML value, add as Swissprot as we do in
            # the equivalent function in ProteinObject.py 
            elif element.nodeName == 'name':
                accList.append((str(element.childNodes[0].nodeValue), "Swissprot-Locus"))
                
            # ________________________________________________________________
            # Look for EMBL protein sequences
            elif element.nodeType == element.ELEMENT_NODE and element.getAttribute("type") == "EMBL":
                
                # EMBL entires need to be further searched for their protein sequence ID values
                for subElement in element.childNodes:
                    if subElement.nodeType == subElement.ELEMENT_NODE and subElement.getAttribute('type') == "protein sequence ID":
                        tempID = subElement.getAttribute('value')
                        accList.append((str(tempID), ProteinParser.ID_type(tempID)[1]))

                        
            # ________________________________________________________________
            # Finally also look for refseq and IPI are a little accession valyes
            elif element.nodeType == element.ELEMENT_NODE and element.getAttribute("type") in ("IPI", "RefSeq"):
                tempID = element.getAttribute('id')
                accList.append((str(tempID), ProteinParser.ID_type(tempID)[1]))

        return list(set(accList))

#-------------------------------------------------------
# PRIVATE FUNCTION
#--------------------------------------------------------
# Internal function which gets the species from which
# this protein is derived
#

    def _getSpecies(self, domObject, ID):
        elementList = domObject.childNodes
        
        for element in elementList:
            if element.nodeName == "organism":
                for subElement in element.childNodes:
                    if subElement.nodeType == subElement.ELEMENT_NODE and subElement.getAttribute("type") == "scientific":
                        return (subElement.childNodes[0].nodeValue)


#-------------------------------------------------------
# PRIVATE FUNCTION
#--------------------------------------------------------
# Internal function which gets the species from which
# this protein is derived
#

    def _getHost(self, domObject, ID):
        
        scientificName = ""
        commonName = ""
        
        hostnames = domObject.getElementsByTagName("organismHost")

        if hostnames: 
            for subElement in hostnames[0].childNodes:
                if subElement.nodeType == subElement.ELEMENT_NODE and subElement.getAttribute("type") == "scientific":
                    scientificName = subElement.childNodes[0].nodeValue
                if subElement.nodeType == subElement.ELEMENT_NODE and subElement.getAttribute("type") == "common":
                    commonName = subElement.childNodes[0].nodeValue
        else:
            return "N/A"

        if scientificName == "" and commonName == "":
            raise UniprotAPIException("Error - unable to find host organism name despite 'organismHost' tags in ID " + ID)
        else:
            return str(scientificName + " (" + commonName + ")") 


#--------------------------------------------------------
# PRIVATE FUNCTION
#--------------------------------------------------------
# Internal function which gets a list of dictionaries which
# hold the pfam domains. These domains are built by querying
# Pfam database directly because domain information is not
# stored on 
#

    def _getDomains(self, sequence, ID):

        #===========================================================
        # internal/local function - avoid
        # class-space polution
        def getPfamDOM(ID):
            handle = self.Network.PfamNetworkRequest(ID, False)
            
            try:
                domObject = parseString(handle.read())
            except Exception, e:
                print e 
                raise UniprotAPIException("Error when converting Pfam XML to DOM object")
            
            return domObject
        #===========================================================
        
        domainList = []
        tempDict = {}
        
        try:
            PfamDomObject = getPfamDOM(ID)
            self._isValidXML(PfamDomObject)
        except UniprotAPIException, e:
            print e
            raise UniprotAPIException("Error when accessing Pfam database with accession " + ID)

        elementList = PfamDomObject.getElementsByTagName("entry")[0].childNodes

        for element in elementList:
            # Ensure the Pfam sequence and our Uniprot sequence match, or the domain locations
            # will miss align!
            if element.nodeName == 'sequence':

                try:
                    
                    if not len(str(element.childNodes[0].nodeValue)) == len(sequence):
                        print "Length difference between sequences"
                        print "Pfam sequence (" + str(len(element.childNodes[0].nodeValue)) +") = " + str(element.childNodes[0].nodeValue).lower()
                        print "Uniprot sequence ("  + str(len(sequence)) +") = " + str(sequence).lower()
                    
                        raise UniprotAPIException("Pfam and Uniprot sequences fail to match for accession " + ID)
                except Exception, e:
                    print e
                    # print element.toxml()
                    return element
            elif element.nodeName == 'matches':
                for match in element.childNodes:
                    if match.nodeName == 'match':
                        for location in match.childNodes:
                            if location.nodeName == 'location':
                                tempDict["start"] = str(location.getAttribute("start"))
                                tempDict["stop"] = str(location.getAttribute("end"))
                                tempDict["type"] = "pfam"
                                tempDict["accession"] = ID
                                tempDict["label"] = str(match.getAttribute("id"))
                                domainList.append(tempDict)
                                tempDict = {}
        return domainList
#--------------------------------------------------------
# PRIVATE FUNCTION
#--------------------------------------------------------
# Internal function which gets the (ordered) taxonomy list
# associated with our protein of interest
#

    def _getTaxonomy(self, domObject, ID):
        elementList = domObject.childNodes

        taxon = []

        for element in elementList:
            if element.nodeName == "organism":
                for subElement in element.childNodes:
                    if subElement.nodeName == "lineage":
                        for taxonVal in subElement.childNodes:
                            if taxonVal.nodeName == "taxon":
                                taxon.append(taxonVal.childNodes[0].nodeValue)

        return taxon
    
#--------------------------------------------------------
# PRIVATE FUNCTION
#--------------------------------------------------------
# Internal function to get the isoform sequences associated
# with a specific protein. Rather than re-constructing from 
# annotations as the NCBI isoforms are done, here we simply 
# make a batch request to UniProt for the various isoform 
# sequences. For now that's how this will stay, although this
# does mean we have to make an extra networking request.
#
# It may be worthwhile to look into an annotation parsing
# approach at a later date, but for now this provides an ideal
# mechanism to ensure that the annotation based algorithm works
# as expected (i.e. provides an independently produced reference
# isoform sequence)
#
#
    def _getIsoforms(self, domObject, ID):
        isoforms = {}
        isoformIDs = []
        isoformDomList = domObject.getElementsByTagName("isoform")
        
        # initially we construct a list of the isoform IDs found associated 
        # with the XML record
        for isoformDom in isoformDomList:
            for subElement in isoformDom.childNodes:
                if subElement.nodeName == "id":
                    isoformIDs.append(subElement.childNodes[0].nodeValue)

        # Then, assuming there is at least one isoform, we query the UniProt
        # servers for the sequences associated with those isoforms
        if len(isoformIDs) > 0:
            isoformRaw = self.Network.UniProtBatchIsoformNetworkRequest(isoformIDs, 'fasta')

            if isoformRaw == -1:
                raise UniprotAPIException("Unable to carry out isoform sequence lookup through UniProt for accession " + ID)
            
            # we can then parse the returned raw data to create a structured
            # Biopython FASTA list
            isoformList = list(SeqIO.parse(StringIO.StringIO(isoformRaw), 'fasta'))

            for record in isoformList:
                
                # get the isoform ID (i.e. in the format [A-Z]XXXXX-[0-9]*)
                isoID = str(record.id[record.id.find("|")+1:record.id.rfind("|")])
                
                # First try and find a name after the word, "isoform" in the description field.
                # Originally I pulled the number after the dash as the isoform number (e.g. Q9NP78-5). However, it turns
                # out this number doesn't *necessarily* point to it's corresponding isoform, case in point, Q9NP78-5 actually
                # refers to isoform 4. Unhelpful much.
                #
                # If this extraction method fails we do go and pull the accession reference and use that as the name
                # - it's right, but potentially misleading!
                
                try:
                    boundary = str(record.description).find("Isoform")+8
                    
                    if boundary == 7:
                        # If we've actually pulled down a non-isoform sequence (no "Isoform" in the name)
                        # then let's just skip over it. Note boundary would == 7 because a failure
                        # to match = -1 and then we add 8!
                        continue

                    endBoundary = record.description[boundary:].find("of") 
                    
                    # It seems that literally every isoform is defined by, "Isoform <name> of" ...
                    # so we just extract everything in the <name> portion to use as the name key.
                    # We may have to add additional rules here.
                    #
                    # Weirdly, using .find gives a much cleaner and more robust solution
                    # than a regex, because inevitably someone names an isoform something
                    # that breaks your regex. Hopefully no one will name the isoform, 
                    # "Isoform"
                    
                    isoformName = str(record.description[boundary:endBoundary+boundary-1])
                    
                    
                # This is a bit bad (catching every exception) but means if any part of the above
                # process goes wrong we default to extracting from the ID    
                except Exception, e:
                    print e 
                    print record.description
                    print "Error when parsing isoform name using description (above), falling back to identifiers..."
                    isoformName = (record.id[record.id.find("|")+1:record.id.rfind("|")])

                
                isoforms[isoID] = [isoformName, record.seq.tostring().lower()]

            return isoforms
                
        else:
            return {}


            
#--------------------------------------------------------
# PRIVATE FUNCTION
#--------------------------------------------------------
# ERROR HANDLING FUNCTIONS
    
    # does the  $nodelist have more than 1 $element tag?    
    def _neq1(self, nodelist, element, ID ):
        if nodelist.length > 1:
            raise UniprotAPIException("Multiple '" + element + "' tags found in UniProt XML string associated with accession number " + ID ) 
        if nodelist.length < 1:
            raise UniprotAPIException("No '" + element + "' tags found in UniProt XML string associated with accession number " + ID ) 

    # does the $nodelist have more than 1 $element tag - show explicit message
    def _neq1_explicit(self, nodelist, message):
        if nodelist.length != 1:
            raise UniprotAPIException(message) 

    # is a string empty - show explicit message
    def _isEmpty(self, string, message):
        if string == "":
            raise UniprotAPIException(message)

    def _isValidXML(self, domObject):
        if not len(domObject.getElementsByTagName("entry")) == 1:
            raise UniprotAPIException("Invalid XML - number of 'entry' tagged XML regions != 1")
            




        

        


            
        
            
            

        



        





import unittest
import urllib
import urllib2
from Bio import Entrez

from geeneus.backend import Networking

class TestNetworkingFunctions(unittest.TestCase):

    # Build manager object for all tests here
    def setUp(self):
        pass

    def test_internet_is_working_up(self):
        try:
            self.assertEquals(200, urllib.urlopen("http://www.google.com").getcode())
        except IOError:
            self.fail("Unable to connect to internet")

    def test_NCBI_is_up(self):
        try:
            self.assertEquals(200, urllib.urlopen("http://eutils.ncbi.nlm.nih.gov/entrez/eutils/").getcode())
        except IOError:
            self.fail("Unable to connect to the eutils base connection")

    def test_ENTREZ_efetch(self):
        Entrez.email = "alex.holehouse@gmail.com"  
        try:
            handle = Entrez.efetch(db="protein", id="NP_005566",  retmode="xml")
        except urllib2.HTTPError:
            self.fail("ENTREZ efetch to the protein database doesn't seem to be work")
            return




# Contains networking functionality  (private)
#
# Abstracts away any interaction with the eUtil tools from the user, and deals with network errors or other problems.
#
# NOTE:
# timeout code based on code from http://pguides.net/python-tutorial/python-timeout-a-function/
# -
# Copyright 2012 - 2015 by Alex Holehouse - see LICENSE for more info
# Contact at alex.holehouse@wustl.edu

import datetime
import sys
import signal
import time
import urllib2
import StringIO
import requests
import ProteinParser
from httplib import BadStatusLine
from Bio import Entrez

#--------------------------------------------------------
# Global networking timeout limits are in ProteinParser
# for fine tuning

######################################################### 
######################################################### 
# Exception class for timeouts
#
class TimeoutException(Exception):
    pass



######################################################### 
######################################################### 
# Main class to handle all the networking shinanigans
#
class Networking:    
    
    TIMEOUT=20
    def __init__(self, timeout):
        self.lastDatabaseCall = datetime.datetime.now()
        
#--------------------------------------------------------
#
#--------------------------------------------------------
# Function to decorate other functions with a timeout. If the timeout is reached, causes the
# decorated function to return a -1 along with a printed warning
#    
    def timeout(timeout_time, default):
        def timeout_function(f):
            def f2(*args):

                def timeout_handler(signum, frame):
                    raise TimeoutException()

                handler = signal.signal(signal.SIGALRM, timeout_handler)
                signal.alarm(timeout_time)
                            
                try:
                    retval = f(*args)
                except TimeoutException:
                    print "\nWarning: Timeout reached after {time} seconds\n".format(time=timeout_time)
                    return default
                finally:
                    signal.signal(signal.SIGALRM, handler)
                    signal.alarm(0)
                return retval
            return f2
        return timeout_function
    

#--------------------------------------------------------
#
#--------------------------------------------------------
# function to ensure we stay with NCBI's query limit of no more than 3 per second
#
#
    def stay_within_limits(self):
        if (datetime.datetime.now() - self.lastDatabaseCall).seconds  < 1:
            if (datetime.datetime.now() - self.lastDatabaseCall).microseconds < 400000:
                time.sleep(0.5)
            
        self.lastDatabaseCall = datetime.datetime.now()

#--------------------------------------------------------
#
#--------------------------------------------------------
# efetch nucleotide sequence
#
# Function to get a live handle with nucleotide xml data. All networking
# issue should be dealt with here and abstracted totally from the user
# Decorator must decorate this function (not efetchGeneral) to avoid keyword
# conflicts
#
# A paired __internal_efNT() and efetchNucleotide() set of functions are used
# to allow decoration of one with a timeout, where the efetchNucleotide() can 
# print an error message on -1 return from EITHER the efetchGeneral function,
# or from the timeout decorator itself
#
    @timeout(TIMEOUT, -1)
    def __internal_efNT(self, GI, start, end, strand_val):
        return self.eUtilsGeneral({'function':Entrez.efetch,'db':"nucleotide", 'id':GI, 'seq_start':start, 'seq_stop':end, 'rettype':"fasta", 'strand':strand_val})
      
    def efetchNucleotide(self, GI, start, end, strand_val):
        self.stay_within_limits()
        handle = self.__internal_efNT(GI, start, end, stand_val)
    
        if (handle == -1):
            print "[NCBI]: Networking Error: Problem getting Nucleotide data for GI|{gi}".format(gi=GI)
            return -1
        else:
            return handle
#--------------------------------------------------------
#
#--------------------------------------------------------
#  efetch gene record
#
# Function to get a live handle with gene xml data. All networking
# issue should be dealt with here and abstracted totally from the user
# Decorator must decorate this function (not efetchGeneral) to avoid keyword
# conflicts
#
    @timeout(TIMEOUT, -1)
    def __internal_efG(self, GeneID):
        return self.eUtilsGeneral({'function':Entrez.efetch,'db':"gene", 'id':GeneID, 'rettype':"gene_table", 'retmode':"xml"})

    def efetchGene(self, GeneID):
        self.stay_within_limits()
        handle = self.__internal_efG(GeneID)
        if (handle == -1):
            print "[NCBI]: Network Error: Problem getting gene  data for ID: {GID}".format(GID=GeneID)
            return -1
        else:
            return handle

#--------------------------------------------------------
#
#--------------------------------------------------------
# efetch protein record
#
# Function to get a live handle with protein xml data. All networking
# issue should be dealt with here and abstracted totally from the user
# Decorator must decorate this function (not efetchGeneral) to avoid keyword
# conflicts
#
    @timeout(TIMEOUT, -1)
    def __internal_efP(self, ProteinID):
        return self.eUtilsGeneral({'function':Entrez.efetch,'db':"protein", 'id':ProteinID, 'retmode':"xml"})
       
    def efetchProtein(self, ProteinID):
        self.stay_within_limits()
        handle = self.__internal_efP(ProteinID)
        if (handle == -1):
            print "[NCBI]: Networking Error: Problem getting protein data for ID(s): {PID}".format(PID=ProteinID)
            return -1
        else:
            return handle

#--------------------------------------------------------
#
#--------------------------------------------------------
# epost protein list (not really used, but here incase we implement GI based
# asynchronous fetching in the future)
#
# Function to get post a list of IDs to the NCBI server
# for asynchronous processing. As of 23 Oct 2012 this is
# not being used, but is kept in case we add asynchronous
# epost based features in the future
#
    @timeout(TIMEOUT, -1)
    def __internal_epP(self, ProteinIDList):
        return self.eUtilsGeneral({'function':Entrez.epost,'db':"protein", 'id':",".join(ProteinIDList)})

    def epostProtein(self, ProteinIDList):
        self.stay_within_limits()
        handle = self.__internal_epP(ProteinIDList)
        if (handle == -1):
            print "[NCBI]: Networking Error: Problem ePosting ID(s): {PID}".format(PID=ProteinIDList)
            return -1
        else:
            return handle


#--------------------------------------------------------
#
#--------------------------------------------------------
# esearch
#
# Function to query the protein database for the $passedTerm
# term 
#

    @timeout(TIMEOUT, -1)
    def __internal_esP(self, passedTerm):
        return self.eUtilsGeneral({'function':Entrez.esearch,'db':"protein", 'term':passedTerm})

    def esearchProtein(self, term):
        self.stay_within_limits()
        print "looking_up ..." + str(term)
        handle = self.__internal_esP(term)
        if (handle == -1):
            print "[NCBI]: Networking Error: Problem eSearching for term: {PID}".format(PID=term)
            return -1
        else:
            return handle

#--------------------------------------------------------
#
#--------------------------------------------------------
# Generic function to make some connection to the NCBI database
#
# While it would make more sense to check we're staying within the
# NCBI limits here (self.stay_within_limits()) this would mess up
# the multithreading used for the timeout decorator, so we use
# stay_within_limits() before we activate the timeout.
#
#
#
    def eUtilsGeneral(self, inputDictionary):
        function = inputDictionary["function"]
        
        del inputDictionary["function"]
        try:
            handle = function(**inputDictionary)
            
        # Below we have error handlers for the various types of errors you might get
        # Additional edge cases can be added here as necessary (think of this like
        # an except switch/case statement
        except (urllib2.HTTPError), err:
            print "[NCBI]: HTTP error({0}): {1}".format(err.code, err.reason)
            return -1 
        except BadStatusLine, err:
            print "[NCBI]: httplib error - " + str(err)
            print "[NCBI]: This is a super rare error! Congratualtions?"
            return -1
        except urllib2.URLError, err:
            try:
                print "[NCBI]: URLError error({0}): {1}".format(err.code, err.reason)
            except AttributeError, err:
                print "[NCBI]: Corrupted urllib2.URLError raised"
                return -1
            return -1
        return handle



###############################################################################################
##                             UNIPROT NETWORKING FUNCTIONS                                  ##
###############################################################################################

#--------------------------------------------------------
#
#--------------------------------------------------------
# UniProtNetworkRequest
#
# Public, single accession Networking function which accesses
# the UniProt servers. Returns either -1 if something
# goes wrong, or the live handle if all is well. Note it is possible
# for the handle to be corrupt, so should be wrapped in try/except
# block when read or parsed.

    def UniProtNetworkRequest(self, accessionID):
        baseURL = 'http://www.uniprot.org/uniprot/'
        queryString = baseURL+str(accessionID)+'.xml'
        
        # probably good to set some kind of limit
        self.stay_within_limits()
        
        return self.__internal_UniprotNR(queryString)
    

    @timeout(TIMEOUT, -1)
    def __internal_UniprotNR(self, queryString):
        return self.__uniprotNetworkingCall(queryString)

    def __uniprotNetworkingCall(self, queryString):
        try:
            handle = urllib2.urlopen(queryString)
        except urllib2.URLError, err:
            try:
                print "[UniProt]: URLError error({0}): {1}".format(err.code, err.reason)
            except AttributeError, err:
                print "Corrupted urllib2.URLError raised"
                return -1
            return -1
        return handle


    def UniProtIsoformNetworkRequest(self, accessionID):
        baseURL = 'http://www.uniprot.org/uniprot/'
        queryString = baseURL+str(accessionID)+'.fasta'
        
        # probably good to set some kind of limit
        self.stay_within_limits()
        
        return self.__internal_UniprotNR(queryString)

    def UniProtBatchIsoformNetworkRequest(self, accessionList, returnFormat):
            
        # use requests to make http-POST more friendly, and StringIO to
        # imitate a file and avoid fileIO bottle neck

        try:
            r = requests.post('http://www.uniprot.org/batch/', files={'file':StringIO.StringIO(' '.join(accessionList))}, params={'format':returnFormat})
        except:
            print "[UNIPROT]: Networking error when batch querying UniProt for isoforms"
            return -1


        retrycounter = 0;

        while 'Retry-After' in r.headers:
            
            if retrycounter == 3:
                print "[UNIPROT] Unable to batch get isoform data... Server request failed"
                return -1
            
            t = int(r.headers['Retry-After'])
            print '[UNIPROT] Initial batch request failed. Waiting and retrying (' + str(retrycounter+1)+" of 3)" 
            time.sleep(t)
            r = requests.get(r.url)
            retrycounter = retrycounter+1

        return r.text

        
    
    

###############################################################################################
## PFAM NETWORKING FUNCTIONS
###############################################################################################

    @timeout(TIMEOUT, -1)
    def __internal_PfamNR(self, queryString):
        
        try:
            handle = urllib2.urlopen(queryString)
        except urllib2.URLError, err:
            try:
                print "[Pfam]: URLError error({0}): {1}".format(err.code, err.reason)
            except AttributeError, err:
                print "Corrupted urllib2.URLError raised"
                return -1
            return -1
        return handle
       

    def PfamNetworkRequest(self, accessionID, limit=True):
        baseURL = "http://pfam.sanger.ac.uk/protein?output=xml&acc="
        queryString = baseURL+str(accessionID)
        
        # probably good to set some kind of limit
        self.stay_within_limits()
        
        return self.__internal_PfamNR(queryString)

            

    



import unittest
import time
from geeneus import backend
import random

class TestProteinParserFunctions(unittest.TestCase):

    def setUp(self):
        
        # build parser object with cache turned off
        self.ParserObject = backend.ProteinParser.ProteinRequestParser("alex.holehouse@gmail.com", cache=True, retry=4)

        self.assertTrue(not self.ParserObject.error()) 

        self.seqAAB29246 = "msepagdvrqnpcgskacrrlfgpvdseqlsrdcdalmagciqearerwnfdfvtetplegdfawervrglglpklylptgprrgrdelgggrrpgtspallqgtaeedhvdlslsctlvprsgeqaegspggpgdsqgrkrrqtsmtdfyhskrrlifskrkp"
        self.seq1842230 = "meepqsdlsielplsqetfsdlwkllppnnvlstlpssdsieelflsenvtgwledsggalqgvaaaaastaedpvtetpapvasapatpwplsssvpsyktfqgdygfrlgflhsgtaksvtctyspslnklfcqlaktcpvqlwvnstpppgtrvramaiykklqymtevvrrcphherssegdslappqhlirvegnlhaeylddkqtfrhsvvvpyeppevgsdcttihynymcnsscmggmnrrpiltiitledpsgnllgrnsfevricacpgrdrrteeknfqkkgepcpelppksakralptntssspppkkktldgeyftlkirgherfkmfqelnealelkdaqaskgsedngahssylkskkgqsasrlkklmikregpdsd"

    def test_get_protein_sequence(self):
        self.assertEqual(self.seqAAB29246, self.ParserObject.get_sequence("AAB29246"))

        
    def test_get_fake_protein_sequence(self):
        print"\nYou should see a warning regarding 'FAKETEST' below"
        self.assertEqual(None, self.ParserObject.get_sequence("FAKETEST"))


    def test_get_batch_protein_sequence(self):
        print"\nYou should see a warning regarding 'FAKETEST' below"
        IDLIST = ["AAB29246", "FAKETEST", "1842230"]
        outlist = self.ParserObject.batchFetch(self.ParserObject.get_sequence, IDLIST)
            
        self.assertEquals(outlist[IDLIST[0]], self.seqAAB29246)
        self.assertEquals(outlist[IDLIST[1]], None)
        self.assertEquals(outlist[IDLIST[2]], self.seq1842230)

    def test_massive_batch_request(self):
        self.ParserObject.purge_data_store()

        
        # note all caps because the caps unifier is at a higher level
        IDLIST = ["AAB29246", "1842230", "Q13480","NP_997006","P16144","NP_000204","NP_001005619","GI|61742777","Q9UIQ6","NP_005566","NP_787116","GI|8923710","Q9UKY7","NP_060018","GI|20986529","NP_002736","GI|4503787","P42685","NP_002022", "GI|20562757","B0LPG3","B3KR49","P27361","Q7Z3H5","NP_001035145","NP_002737","XP_055766","IPI00018195","GI|20357552","Q14247","Q53HG7","Q76MU0","NP_005222","IPI00029601","GI|21361602","A8K6H9","Q96QD8","Q96QD8-1","Q9HAV3","NP_061849","IPI00410034","P49023","P49023-1","PAXI_HUMAN","NP_002850","IPI00335634","C9JH84","NP_065939","IPI00001632","IPI00641889","GI|4759082","A8K2W3","O95810","SDPR_HUMAN","NP_004648","Q9YH08","GI|9257199","Q9UQB8","NP_001138360","NP_059344","NP_059345","GI|37551295","GI|50843820","GI|51468542","Q5T5P2","NP_001091970","NP_062536","XP_166112","XP_495798","GI|27886584","Q14289","NP_004094","NP_775266","NP_775268","GI|4505055","A8K379","P07948","P07948-1","NP_001104567","NP_002341","IPI00298625","GI|50845388","P07355-2","NP_001002858","IPI00418169","GI|52426745","P22681","NP_005179","IPI00027269","Q6ZVM7","Q6ZVM7-1","NP_001076437","IPI00446294","P30530","P30530-1","NP_068713","IPI00296992","Q01804","Q01804-1","NP_059963","IPI00399254","Q86YV5","NP_001074295","IPI00166578","IPI00739386","GI|15451856","A9XTE5","Q03135","Q03135-1","Q2TNI1","NP_001744","IPI00009236","GI|5803137","P98179","NP_006734","IPI00024320","GI|4758248","P98172","NP_004420","IPI00024307","GI|21361181","P05023","NP_000692","NP_001153705","GI|16936528","P24941","NP_001789","GI|4506403","A8K556","B3KV45","Q8NFJ5","NP_003970","IPI00022624"]
        
        outlist = self.ParserObject.batchFetch(self.ParserObject.get_sequence, IDLIST)

        counter = 0
        for i in IDLIST:
            print counter
            print i
            print outlist[i]
            counter = counter+1


        # generate 10 numbe
        mylist = list(xrange(133))
        selectList = random.sample(mylist, 20)
        testList = []
        for i in selectList:
            print "Accession ID: " + IDLIST[i]
            print self.ParserObject.get_taxonomy(IDLIST[i])
            print self.ParserObject.get_other_accessions(IDLIST[i])
            print self.ParserObject.get_species(IDLIST[i])
            print self.ParserObject.get_domains(IDLIST[i])
            print self.ParserObject.get_gene_name(IDLIST[i])
            print self.ParserObject.get_geneID(IDLIST[i])
            print self.ParserObject.get_record_version(IDLIST[i])
            print self.ParserObject.get_other_accessions(IDLIST[i])
            
          
        self.assertEquals(104, self.ParserObject.datastore_size())
        

    def test_get_variants(self):
        variants = self.ParserObject.get_variants("P16144")
        if len(variants) > 0:
            self.assertEquals(38, variants[0]['location'])
        else:
            print "If we're connected to the internet this test has failed!"


    def test_get_number_of_items(self):
        time.sleep(2) # so we don't go over threshold values
        testMan = backend.ProteinParser.ProteinRequestParser("alex.holehouse@gmail.com", cache=True, retry=4)
        
        self.assertEquals(0, testMan.datastore_size())
        testMan.get_protein_name("P05023")
        self.assertEquals(1, testMan.datastore_size())
        del testMan
        

    def test_purge(self):
        self.ParserObject.purge_data_store()
        self.assertEquals(0, self.ParserObject.datastore_size())
        



import unittest
from geeneus.backend import UniprotAPI


class TestUniprotAPIFunctions(unittest.TestCase):

    # Build manager object for all tests here
    def setUp(self):
        
        self.listOfIDs = ["A0AVF1", "A2A3L6", "B0I1T2"]
        
        # set listOfIDs to a big list if we want to put UniProt server through its paces
        if True:
            self.listOfIDs = ["A0AVF1", "A2A3L6", "B0I1T2", "C9JDP6", "E2RYF6", "O00116", "P01889", "P07951", "Q15714", "Q16549", "Q9Y6Z4"]


        self.UAPI = UniprotAPI.UniprotAPI()
        
    def test_rapidfire(self):

        temp = {}
        
        for ID in self.listOfIDs:
            print "ID = " + ID
            self.UAPI.getProteinObjectFromUniProt(temp, ID)
            
            
            temp[ID].get_geneID()
            temp[ID].get_protein_sequence()
            temp[ID].get_variants()
            temp[ID].get_protein_sequence_length()
            temp[ID].get_domains()
            temp[ID].get_other_accessions()
            temp[ID].get_species()
            temp[ID].get_taxonomy()
            temp[ID].get_host()
			
            temp[ID].exists()
            temp[ID].error()
            
    def test_get_name(self):

        datastore = {}
        
        self.UAPI.getProteinObjectFromUniProt(datastore, "Q9Y6Z4")
        print "\nBAD_ID should yeild a gracefull error/ignore"
        self.UAPI.getProteinObjectFromUniProt(datastore, "BAD_ID")
        self.UAPI.getProteinObjectFromUniProt(datastore, "P01889")

        self.assertEqual(datastore["Q9Y6Z4"].get_protein_name(), "Putative uncharacterized protein KIF25-AS1")
        self.assertEqual(datastore["P01889"].get_protein_name(), "HLA class I histocompatibility antigen, B-7 alpha chain")


    def test_get_protein_sequence(self):
        pass


    def test_get_variants(self):
        pass

    def test_get_geneID(self):
        pass



# Copyright 2012 by Alex Holehouse - see LICENSE for more info
# Contact at alex.holehouse@wustl.edu

""" This is test folder for running unit tests.
"""
import sys
import os
# Add the parent directory (which holds the geeneus package)
sys.path.insert(0,os.path.abspath(__file__+"/../../../"))

import Proteome_tests
import Genome_tests
import GeneObject_tests
import GeneParser_tests
import ProteinObject_tests
import ProteinParser_tests
import Networking_tests
import UniprotAPI_tests
import Isoform_algorithm_tests



# Copyright 2012-2015 by Alex Holehouse - see LICENSE for more info
# Contact at alex.holehouse@wustl.edu

""" This is the backend module which facilitates data structures, utility functions and other stuff the user shouldn't deal with. Everything here isn't designed to be used as part of the geeneus API - the only functions which should be used there are the public ones in the geeneus.
"""
import GeneObject
import GeneParser
import ProteinObject
import ProteinParser



# Public facing API for accessing gene information
#
# Copyright 2012-2015 by Alex Holehouse - see LICENSE for more info
# Contact at alex.holehouse@wustl.edu
import Bio.Entrez
import backend

class GeneManager:
    """ Manager object used to access relevant information via the below methods. The *Manager classes are the only public facing classes in Geeneus, and are the only ones which should be used to access remote information."""

    def __init__(self, email, cache=True, retry=0):
        """Returns a fully formed manager object which can be queried by the other 
        functions in this class. 

        email    Must be a valid email, as is required by the NCBI servers. For more 
                 information on NCBI usage guidelines please see 
                 [http://www.ncbi.nlm.nih.gov/books/NBK25497/].

        cache    Determines if the manager object should cache requests in memory, or 
                 the NCBI database should be queried every time a request is made. 

        retry    The number of times the networking utilities will retry on a failed 
                 connection

        timeout  The number of seconds the networking utilities wait after making a 
                 request before deciding that request has failed
        """
        
        self.datastore = backend.GeneParser.GeneRequestParser(email, cache, retry)

        if self.datastore.error():
            self.error_status = True
        else:
            self.error_status = False
    
    def get_gene_sequence(self, ID):
        """Returns the DNA sequence associated with this accession number"""
        print "+-------------------------------------------------------------------------------------+"
        print "| This function will return a gene sequence. Please wait for 0.1.4 for functionality  |"
        print "+-------------------------------------------------------------------------------------+"
        
        #return self.datastore.get_consensus_sequence(ID)
        
    def get_gene_coding_sequence(self, ID):
        print "+----------------------------------------------------------------------+"
        print "| This function will return a dictionary of isoform-sequence pairings  |"
        print "+----------------------------------------------------------------------+"

    def get_gene_SNP(self, ID):
        print "+--------------------------------------------------------------------------------------------+"
        print "| This function will return a dictionary of SNP-number:Dictionary with SNPs for the gene ID  |"
        print "+--------------------------------------------------------------------------------------------+"
    
    def get_raw_xml(self, ID):
        return self.datastore.get_raw_xml(ID)



# Provides an object-based API for a gene (private)
#
# Copyright 2012 - 2015 by Alex Holehouse - see LICENSE for more info
# Contact at alex.holehouse@wustl.edu

from Bio import Entrez, Seq
from Bio import SeqIO
from Bio.Alphabet import IUPAC

# object attributes
#

class GeneObject:

#--------------------------------------------------------
#
#--------------------------------------------------------
    def __init__(self, ID, genexml):
        
        self.__set_default_attributes(error=False)

        # check for error, and return with error flag = True
        if genexml == -1:        
            self._error = True
            return
        
        self.ID = ID
        self.raw_xml = genexml

        # check xml is a real, correct single gene
        if not self.__xml_is_OK(genexml):
            return
        
        # if we've got here we have a non-errorful, real
        # xml structure to deal with, so the object exists
        self._exists = True
        
        # Building the full sequence requires an additional network call to pull down
        # the nucleotide data. Should this fail, the __build_full_sequence() function
        # will return -1, in which case we reset the object status to error and return
        if (self.___build_full_sequence(genexml) == -1):
            self.__reset_to_error()
            return
        
        self.accession = self.__extract_accession(genexml)
        self.version = self.__extract_version(genexml)
        self.GI = self.__extract_GI(genexml)
        self.locus = self.__extract_locus(genexml)

#--------------------------------------------------------
#
#--------------------------------------------------------
# object gettes for attributes

    def get_full_sequence(self):
        return (self.full_sequence, self.full_sequence_c)

    def get_version(self):
        return self.version

    def get_accession(self):
        return self.accession

    def get_versioned_accession(self):
        return self.accession + "."+ self.version

    def get_gi(self):
        return self.GI

    def get_raw_xml(self):
        return self.raw_xml
        
#--------------------------------------------------------
#
#--------------------------------------------------------
# Returns the coding sequence for the gene, as defined by
# the exons
#
    def get_coding_sequence(self, genexml):
        # stuff
        print genexml[0]["Entrezgene_locus"][0]["Gene-commentary_products"][0]["Gene-commentary_genomic-coords"][0]["Seq-loc_mix"]["Seq-loc-mix"]
        return "ok"


#--------------------------------------------------------
#
#--------------------------------------------------------
# Get gene version from XML
#
    def __extract_version(self, genexml):
        return genexml[0]["Entrezgene_locus"][0]["Gene-commentary_version"]

#--------------------------------------------------------
#
#--------------------------------------------------------
# Get gene accession from XML
#
    def __extract_accession(self, genexml):
        return genexml[0]["Entrezgene_locus"][0]["Gene-commentary_accession"]

#--------------------------------------------------------
#
#--------------------------------------------------------
# get gene GI from XML
#
    def __extrat_GI(self, genexml):
        gene_seq_dict = genexml[0]["Entrezgene_locus"][0]["Gene-commentary_seqs"][0]["Seq-loc_int"]["Seq-interval"]
        return int(gene_seq_dict["Seq-interval_id"]["Seq-id"]["Seq-id_gi"])
#--------------------------------------------------------
#
#--------------------------------------------------------
# get gene locus from XML 
#
    def __extract_locus(self, genexml):
        return genexml[0]["Entrezgene_location"][0]["Maps_display-str"]
    
    
    def __set_default_attributes(self, error):
        self.full_sequence = ""
        self.full_sequence_c = ""
        self._exists = False
    
    # if there's an error...
        if (error):
            self._error = True
        else:
            self._error = False
            self.full_sequence_length = 0
            self.accession = ""
            self.version = ""
            self.GI = ""
            self.locus = ""

#--------------------------------------------------------
#
#--------------------------------------------------------
# Internal function which takes the genexml, and extracts the full
# coding sequence. If anything goes wrong while pulling nucleotide
# data will return -1, else returns nothing, but changes the state 
# of the objects self.full_sequence[_c] variables behind the 
# scenes #antifunctionalprogramming.
#
    def ___build_full_sequence(self, genexml):
        
        gene_seq_dict = genexml[0]["Entrezgene_locus"][0]["Gene-commentary_seqs"][0]["Seq-loc_int"]["Seq-interval"]
        
        # NB - we could get GI from self.GI, but that makes the assumption
        # that we've already run and set self.GI. Best not to do so!
        GI = int(gene_seq_dict["Seq-interval_id"]["Seq-id"]["Seq-id_gi"])
        
        # get positional and locus information (need to add 1!)
        pos_from = int(gene_seq_dict["Seq-interval_from"])+1
        pos_to = int(gene_seq_dict["Seq-interval_to"])+1

        handle = Networking.eFetchNucleotide(GI, pos_from, pos_to, 1) 
        handle_2 = Networking.eFetchNucleotide(GI, pos_from, pos_to, 2) 
        
        # if we get an error, return -1, which is dealt with in __init__ and
        # the object is set to an error state
        if (handle == -1 or handle_2 == -1):
            return -1
        
        #  given we didn't return -1 no errors, occured, so grab sequence. If t
        try:
            self.full_sequence = SeqIO.read(handle, "fasta")
            self.full_sequence_c = SeqIO.read(handle_c, "fasta")

        # TO DO - catch relevant exceptions only!`
        except:
            return -1
        

        
        
        handle.close()
        handle_c.close()

        return

#--------------------------------------------------------
#
#--------------------------------------------------------
#
#

    def __xml_is_OK(self, genexml):
        if len(genexml) > 1:
            print "WARNING[GeneObject.__xml_is_ok()] - GeneXML detected more than one record associated with this GI.\nThis should never happen."

        if len(genexml) == 0:
            return False
        
        



# Public facing API for accessing protein information
#
# Copyright 2012-2015 by Alex Holehouse - see LICENSE for more info
# Contact at alex.holehouse@wustl.edu

import Bio.Entrez
import geeneus.backend

class ProteinManager:
    
    def __init__(self, email, cache=True, retry=0, uniprotShortcut=True):
        """Returns a fully formed manager object which can be queried by the other 
        functions in this class. 

        email           Must be a valid email, as is required by the NCBI servers. For more 
                        information on NCBI usage guidelines please see 
                        [http://www.ncbi.nlm.nih.gov/books/NBK25497/].

        cache           Determines if the manager object should cache requests in memory, or 
                        the NCBI database should be queried every time a request is made. 

        retry           The number of times the networking utilities will retry on a failed 
                        connection

        uniprotShortcut If True then for non NCBI supported UniProtKB/Swissprot accessions
                        we skip an NCBI lookup and go straight for UniProt. If False then we
                        always try NCBI first.
        """

        self.datastore = geeneus.backend.ProteinParser.ProteinRequestParser(email, cache, retry, shortcut=uniprotShortcut)
        if self.datastore.error():
            self.error_status = True
        else:
            self.error_status = False
            

    #----------------------------------------------------------------------------
    # decorators for streamlining ID checking
   
    def convertIDToUpperCase(fn):
        def wrapped(_, a):
            return fn(_, str(a).upper())
        return wrapped

    def convertIDListToUpperCase(fn):
        def wrapped(_, a):
            return fn(_, [str(x).upper() for x in a])
        return wrapped
    #----------------------------------------------------------------------------

    
    # Firstly we have the general pm functions (i.e. those not linked to
    # a specific accession
                  
    def keys(self):
        """ Return a list of the keys in the datastore (note this hides 
        the always present '-1' key
        """
        return self.datastore.keys()


    def accession_classes(self):
        """ Return a list defines the names of the possible
            'other_accession' accession types. 

            Useful for filtering out non-types by doing

               for i in other_accessions_list:
                  if i in pm.accession_classes():
                      some functionalty 
                      
        """

        return(self.datastore.accession_classes())


    @convertIDToUpperCase
    def has_key(self, ID):
        """ Check if the datastore has a protein ID loaded or not. Does
            not preload the ID in question
        """
        return self.datastore.has_key(ID)

    @convertIDToUpperCase
    def error(self, ID):
        """ Determine if the record associated with the ID had an error 
            on download. Does not preload the ID in question, so returns
            True, False or None if no record has been downloaded
        """
        
        if self.datastore.has_key(ID):
            return self.datastore.protein_error(ID)
        else:
            return None
        
    @convertIDToUpperCase
    def exists(self, ID):
        """ Determine if the record associated with the ID exists as 
            a protein record. Does not preload the ID in question, so returns
            True, False or None if no record has been downloaded.
            
            For example, if we had previously tried to download some accession
            and discovered it was not a valid protein accession, then this would
            return False.
        """

        if self.datastore.has_key(ID):
            return self.datastore.protein_exists(ID)
        else:
            return None

    @convertIDToUpperCase
    def get_record_source(self, ID):
        """ Returns the name of the database used to obtain the record
            in question. If the record has not yet been obtained this
            triggers its fetching/downloading
        """
        return self.datastore.get_protein_source(ID)

    @convertIDToUpperCase
    def get_record_creation_date(self, ID):
        """ Returns the sequence creation date [NB - previously
            this function was going to be renamed to improve clarity,
            but in the interest of not breaking things, it is being
            left as is for the time being]
         """
        return self.datastore.get_creation_date(ID)

    @convertIDToUpperCase
    def get_protein_name(self, ID):
        """ Returns the name of the protein """
        return self.datastore.get_protein_name(ID)

    @convertIDToUpperCase
    def get_record_version(self, ID):
        """ Returns the version of the record (if no standard
            version available just returns 1
        """
        return self.datastore.get_record_version(ID)

    
    @convertIDToUpperCase
    def get_protein_sequence(self, ID):
        """ Return the protein's primary amino acid sequence as a string """
        return self.datastore.get_sequence(ID)

    @convertIDToUpperCase
    def get_raw_xml(self, ID):
        """ Return the raw XML associated with this accession value """
        return self.datastore.get_raw_xml(ID)

    @convertIDToUpperCase
    def get_variants(self, ID):
        """ Return a list of dictionaries, where each dictionary is a mutation 
            dictionary with six keys - 

            location   Position in the primary sequence
            original   Original amino acid 
            mutant     Mutant amino acid
            type       Type of mutation - one of 
            variant    Single term summary showing "Original -> Mutant"
            notes      Annotation notes from download 
         """
            
        return self.datastore.get_variants(ID)

    @convertIDToUpperCase
    def get_geneID(self, ID):
        """ Proteins are also associated with specific genes. This returns 
            the gene ID associated with this accession number. This may be 
            empty if no true gene is associated.
        """
        return self.datastore.get_geneID(ID)

    @convertIDToUpperCase
    def get_gene_name(self, ID):
        """ Get the name of the gene associated with this protein. This may be empty if 
            no true gene is associated.
        """
        return self.datastore.get_gene_name(ID)

    @convertIDToUpperCase
    def get_taxonomy(self, ID):
        """ Get the NCBI defined taxonomy string associated with the organism or
            origin this protein comes from. Returns an ordered list, where the order
            reflects the taxonomy hierarchy displayed on the NCBI website.
        """
        return self.datastore.get_taxonomy(ID)

    @convertIDToUpperCase
    def get_domains(self, ID):
        """ Get a list of dictionary objects, where each dictionary refers to a
            single Pfam domain. These dictionaries have five entires each;
            
            start       Start location in sequence
            stop        Stop location in sequence
            type        Type of domain (always Pfam, at the moment, but allows for
                        future development where other domain types may be of
                        interest
            accession   The (protein) accession value associated with this domain 
            label

            NOTE: While NCBI records hold the Pfam details, if NCBI lookup fails
                  and the system falls back to query UniProt a second query to 
                  the Pfam database must also be made, because UniProt records 
                  only hold references to Pfam domains, not their associated 
                  details. 
        """
        return self.datastore.get_domains(ID)

    @convertIDToUpperCase
    def get_species(self, ID):
        """ Returns the species from which this protein was extracted.
        """
        return self.datastore.get_species(ID)

    @convertIDToUpperCase
    def get_host(self, ID):
        """ Returns the host of the virus if the ID corresponds to a 
            viral protein in the format <scientific name> (<common name>),
            e.g. Homo Sapiens (Human).
            
            If not a viral protein returns the string 'N/A'
        """
        return self.datastore.get_host(ID)

    @convertIDToUpperCase
    def get_other_accessions(self, ID):
        """ Returns a list of tuples which define ("type of accession", "value") 
            for other accessions linked to this accession value.
            
            The possible types are a controlled vocabulary, and we limit the selection
            of other accession values to
            * "Swissprot"
            * "RefSeq"
            * "GI"
            * "PDB"
            * "UniProt"
            * "International Protein Index"
            * "DDBJ"
            * "GenBank"
            * "EMBL"

            If you have an accession value, you can always test its type by calling
            ProteinManagerObject.get_ID_type(ID). If you have an accession value you
            think should be returning a type but isn't, this may reflect an error in
            how Geeneus defines types - a bug report would be greatly appreciated.
        """
                
        return self.datastore.get_other_accessions(ID)
    
    @convertIDToUpperCase
    def get_protein_sequence_length(self, ID):
        """ Returns an integer equal to the length of the protein's primary sequence. 
        """
        return len(self.datastore.get_sequence(ID))

    @convertIDToUpperCase
    def get_ID_type(self, ID):
        """ Returns a two position list, where list[0] is an exit code 
            and list[1] is the name of the type of accession number. 
        """
        return self.datastore.get_ID_type(ID)

    @convertIDToUpperCase
    def get_isoforms(self, ID):
        """ Returns a dictionary, where the keys are isoform names
            and the values are the full isoform sequences associated
            with those values as reconstructed from NCBI records.

            If NCBI lookup fails and we fall back to UniProt, then
            The UniProt isoform sequences will be queried directly,
            but this takes longer as it requires additional network
            requests"
        """
        return self.datastore.get_isoforms(ID)

    @convertIDToUpperCase
    def run_translation(self, Acc):
        """ Translates an alphanumeric accession number to a GI number. """
        return self.datastore.translate_Asc2GI(Acc)

    @convertIDListToUpperCase
    def batch_get_protein_sequence(self, IDList):
        """ Allows for primary sequence retrieval en-masse using a list of IDs
            as input. Returns a dictionary, indexed by accession number, of
            the various sequences. Makes a single request to the NCBI server """
        return self.datastore.batchFetch(self.datastore.get_sequence, IDList)

    @convertIDListToUpperCase
    def batch_get_protein_name(self, IDList):
        """ Get the names of a list of IDs """
        return self.datastore.batchFetch(self.datastore.get_protein_name, IDList)

    @convertIDListToUpperCase
    def batch_get_variants(self, IDList):
        """ Allows variants retrieval en-mass using a list of IDs as input.
            Returns a dictionary of lists, where each dictionary entry is 
            indexed by a protein ID from the list and each list is made up of
            a dictionaries which contian variant information """
        return self.datastore.batchFetch(self.datastore.get_variants, IDList)

    def purge(self):
        """ Wipe the managers memory store. Only relevant if you're worried
            about memory or running a script as a daemon which is constantly
            pulling down data for a one time use """
        self.datastore.purge_data_store()

    def datastore_size(self):
        """ Get the number of items in the internal datastore """
        return self.datastore.datastore_size()

    def load_local_uniprot_xml(self, filename):
        self.datastore.loadfile(filename)

    def save_datastore(self, filename):
        self.datastore.save_datastore(filename)

    def load_datatore(self, filename):
        self.datastore.load_datatore(filename)
    



from distutils.core import setup

setup(
    name='Geeneus',
    version='0.1.9',
    author='Alex Holehouse',
    author_email='alex.holehouse@gmail.com',
    packages=['geeneus', 'geeneus.test', 'geeneus.backend'],
    scripts=[],
    url='http://alexholehouse.github.com/Geeneus/',
    license='LICENSE.txt',
    description='Simple API for NCBI database access',
    long_description=open('README.txt').read(),
    install_requires=[
        "Biopython >= 1.60",
        "requests >= 0.14.2"   
        ],
    )



# Deals with cache and storing of gene data objects, as well as coordinating network access (private)
#
# Copyright 2012 - 2015 by Alex Holehouse - see LICENSE for more info
# Contact at alex.holehouse@wustl.edu

from Bio import Entrez
from Bio.Seq import Seq
from Bio.Alphabet import IUPAC

import GeneObject
import Networking
import Parser as GRP

# RequestParser is a wrapper datastore layer, which facilitates non-redundant access to the
# GeneObjects. It is also where direct communication with the NCBI server occurs, but no xml
# parsing actually happens here.
class GeneRequestParser(GRP.GeneralRequestParser):
    
    def __init__(self, email, cache, retry=0, loud=True):
        try:
            GRP.GeneralRequestParser.__init__(self, email, cache, retry, loud)
            
            self.gene_datastore = {-1 : GeneObject.GeneObject(-1, [])}
            self.gene_translationMap = {-1: -1}
            self.batchableFunctions = [] # empty for now...
            
            # If we get here we're good to go!
            self.error_status = False
           
        except Exception, e:
            print eb
            print "Fatal error when creating GeneRequestParserObject"
            self.error_status = True

            
    #--------------------------------------------------------
    # PUBLIC FUNCTION
    #--------------------------------------------------------
    #
    # Check for an error in the parser object
    # Primary function for querying the gene_datastore in the RequestParser object
    # Functions which are getting information on a Gene based on GeneID should call this
    # and functionality associated with extracting information from the Gene should be
    # added to the GeneObject class.

    # Return value:
    # Will return a GeneObject, or a -1 in the case of an error
    
    def get_gene_object(self, GeneID):
        self._get_object(GeneID, self.gene_datastore, self.Networking.efetchGene, GeneObject.GeneObject)
        return self.gene_datastore[GeneID]
        
    def get_consensus_sequence(self, GeneID):
        gene = self.get_gene_object(GeneID)
        return gene.get_full_sequence()

    def get_raw_xml(self, GeneID):
        gene = self.get_gene_object(GeneID)
        return gene.get_raw_xml()

#--------------------------------------------------------
# PUBLIC FUNCTION
#--------------------------------------------------------
# Function which purges the datastore
#
    def purge_data_store(self):
        del self.gene_datastore
        self.gene_datastore = {-1 : GeneObject.GeneObject([])}

#--------------------------------------------------------
# PUBLIC FUNCTION
#--------------------------------------------------------
# Get the number of items in the datastore
#
    def get_size_of_datastore(self):

        # -1 offset because we have the -1 entry in
        return len(self.Gene_datastore)-1
   
